"""This package contains specific torch-based vector models and the corresponding low-level torch modules."""

# imports required for backward compatibility (with pickled objects)
from .lstnet.lstnet_models import *
from .mlp.mlp_models import *
from .residualffn.residualffn_models import *



import pandas as pd

from ..vector_model import VectorModel


class VectorModelWithSeparateFeatureGeneration:
    def __init__(self, vector_model: VectorModel):
        self.vectorModel = vector_model
        self.featureGen = vector_model.get_feature_generator()
        self.vectorModel.set_feature_generator(None)

    def __str__(self):
        return self.vectorModel.__str__()

    class IntermediateFittingStep:
        def __init__(self, vector_model: VectorModel, x: pd.DataFrame, y: pd.DataFrame):
            self.y = y
            self.x = x
            self.vector_model = vector_model

        def execute(self) -> VectorModel:
            self.vector_model.fit(self.x, self.y)
            return self.vector_model

        def __str__(self):
            return f"{self.__class__.__name__} for {self.vector_model}"

    class PredictFinaliser:
        def __init__(self, vector_model: VectorModel, x: pd.DataFrame):
            self.X = x
            self.vectorModel = vector_model

        def execute(self) -> pd.DataFrame:
            return self.vectorModel.predict(self.X)

        def __str__(self):
            return f"{self.__class__.__name__} for {self.vectorModel}"

    def fit_start(self, x, y) -> 'VectorModelWithSeparateFeatureGeneration.IntermediateFittingStep':
        x = self.featureGen.fit_generate(x, y)
        return self.IntermediateFittingStep(self.vectorModel, x, y)

    def predict_start(self, x: pd.DataFrame):
        x = self.featureGen.generate(x)
        return self.PredictFinaliser(self.vectorModel, x)

    def fit_end(self, vector_model) -> VectorModel:
        vector_model._featureGenerator = self.featureGen
        return vector_model



"""
This module contains sample datasets to facilitate testing and development.
"""
from abc import ABC, abstractmethod

import sklearn.datasets

from sensai.data import InputOutputData
import pandas as pd

from sensai.util.string import ToStringMixin


class DataSet(ToStringMixin, ABC):
    @abstractmethod
    def load_io_data(self) -> InputOutputData:
        pass


class DataSetClassificationIris(DataSet):
    def load_io_data(self) -> InputOutputData:
        iris_data = sklearn.datasets.load_iris()
        input_df = pd.DataFrame(iris_data["data"], columns=iris_data["feature_names"]).reset_index(drop=True)
        output_df = pd.DataFrame({"class": [iris_data["target_names"][idx] for idx in iris_data["target"]]}) \
            .reset_index(drop=True)
        return InputOutputData(input_df, output_df)


class DataSetClassificationTitanicSurvival(DataSet):
    URL = "https://raw.githubusercontent.com/datasciencedojo/datasets/master/titanic.csv"

    COL_INDEX = "PassengerId"
    """
    unique identifier for each passenger
    """
    COL_SURVIVAL = "Survived"
    """
    0 = No, 1 = Yes
    """
    COL_NAME = "Name"
    """
    passenger name
    """
    COL_PASSENGER_CLASS = "Pclass"
    """
    Ticket class as an integer (1 = first, 2 = second, 3 = third)
    """
    COL_SEX = "Sex"
    """
    'male' or 'female'
    """
    COL_AGE_YEARS = "Age"
    """
    age in years (integer)
    """
    COL_SIBLINGS_SPOUSES = "SibSp"
    """
    number of siblings/spouses aboard the Titanic
    """
    COL_PARENTS_CHILDREN = "Parch"
    """
    number of parents/children aboard the Titanic
    """
    COL_FARE_PRICE = "Fare"
    """
    amount of money paid for the ticket
    """
    COL_CABIN = "Cabin"
    """
    the cabin number (if available)
    """
    COL_PORT_EMBARKED = "Embarked"
    """
    port of embarkation ('C' = Cherbourg, 'Q' = Queenstown, 'S' = Southampton)
    """
    COL_TICKET = "Ticket"
    """
    the ticket number
    """
    COLS_METADATA = [COL_NAME, COL_TICKET, COL_CABIN]
    """
    list of columns containing meta-data which are not useful for generalising prediction models
    """

    def __init__(self, drop_metadata_columns: bool = False):
        """
        :param drop_metadata_columns: whether to drop meta-data columns which are not useful for a
            generalising prediction model
        """
        self.drop_metadata_columns = drop_metadata_columns

    def load_io_data(self) -> InputOutputData:
        df = pd.read_csv(self.URL).set_index(self.COL_INDEX, drop=True)
        if self.drop_metadata_columns:
            df.drop(columns=self.COLS_METADATA, inplace=True)
        return InputOutputData.from_data_frame(df, self.COL_SURVIVAL)



from abc import ABC, abstractmethod
import logging
from typing import Any, Union, Optional

import numpy as np
import pandas as pd

from .data_transformation import DFTNormalisation
from .featuregen import FeatureGeneratorFromColumnGenerator
from .util.cache import KeyValueCache


log = logging.getLogger(__name__)


class ColumnGenerator:
    """
    Generates a single column (pd.Series) from an input data frame, which is to have the same index as the input
    """
    def __init__(self, generated_column_name: str):
        """
        :param generated_column_name: the name of the column being generated
        """
        self.generatedColumnName = generated_column_name

    def generate_column(self, df: pd.DataFrame) -> pd.Series:
        """
        Generates a column from the input data frame

        :param df: the input data frame
        :return: the column as a named series, which has the same index as the input
        """
        result = self._generate_column(df)
        if isinstance(result, pd.Series):
            result.name = self.generatedColumnName
        else:
            result = pd.Series(result, index=df.index, name=self.generatedColumnName)
        return result

    @abstractmethod
    def _generate_column(self, df: pd.DataFrame) -> Union[pd.Series, list, np.ndarray]:
        """
        Performs the actual column generation

        :param df: the input data frame
        :return: a list/array of the same length as df or a series with the same index
        """
        pass

    def to_feature_generator(self,
            take_input_column_if_present: bool = False,
            normalisation_rule_template: DFTNormalisation.RuleTemplate = None,
            is_categorical: bool = False):
        """
        Transforms this column generator into a feature generator that can be used as part of a VectorModel.

        :param take_input_column_if_present: if True, then if a column whose name corresponds to the column to generate exists
            in the input data, simply copy it to generate the output (without using the column generator); if False, always
            apply the columnGen to generate the output
        :param is_categorical: whether the resulting column is categorical
        :param normalisation_rule_template: template for a DFTNormalisation for the resulting column.
            This should only be provided if is_categorical is False
        :return:
        """
        return FeatureGeneratorFromColumnGenerator(self,
            take_input_column_if_present=take_input_column_if_present,
            normalisation_rule_template=normalisation_rule_template,
            is_categorical=is_categorical)


class IndexCachedColumnGenerator(ColumnGenerator):
    """
    Decorator for a column generator which adds support for cached column generation where cache keys are given by the input data frame's
    index. Entries not found in the cache are computed by the wrapped column generator.

    The main use case for this class is to add caching to existing ColumnGenerators. For creating a new caching
    ColumnGenerator the use of ColumnGeneratorCachedByIndex is encouraged.
    """

    log = log.getChild(__qualname__)

    def __init__(self, column_generator: ColumnGenerator, cache: KeyValueCache):
        """
        :param column_generator: the column generator with which to generate values for keys not found in the cache
        :param cache: the cache in which to store key-value pairs
        """
        super().__init__(column_generator.generatedColumnName)
        self.columnGenerator = column_generator
        self.cache = cache

    def _generate_column(self, df: pd.DataFrame) -> pd.Series:
        # compute series of cached values
        cache_values = [self.cache.get(nt.Index) for nt in df.itertuples()]
        cache_series = pd.Series(cache_values, dtype=object, index=df.index).dropna()

        # compute missing values (if any) via wrapped generator, storing them in the cache
        missing_values_df = df[~df.index.isin(cache_series.index)]
        self.log.info(f"Retrieved {len(cache_series)} values from the cache, {len(missing_values_df)} still to be computed by "
                      f"{self.columnGenerator}")
        if len(missing_values_df) == 0:
            return cache_series
        else:
            missing_series = self.columnGenerator.generate_column(missing_values_df)
            for key, value in missing_series.iteritems():
                self.cache.set(key, value)
            return pd.concat((cache_series, missing_series))


class ColumnGeneratorCachedByIndex(ColumnGenerator, ABC):
    """
    Base class for column generators, which supports cached column generation, each value being generated independently.
    Cache keys are given by the input data frame's index.
    """

    log = log.getChild(__qualname__)

    def __init__(self, generated_column_name: str, cache: Optional[KeyValueCache], persist_cache=False):
        """
        :param generated_column_name: the name of the column being generated
        :param cache: the cache in which to store key-value pairs. If None, caching will be disabled
        :param persist_cache: whether to persist the cache when pickling
        """
        super().__init__(generated_column_name)
        self.cache = cache
        self.persistCache = persist_cache

    def _generate_column(self, df: pd.DataFrame) -> Union[pd.Series, list, np.ndarray]:
        self.log.info(f"Generating column {self.generatedColumnName} with {self.__class__.__name__}")
        values = []
        cache_hits = 0
        column_length = len(df)
        percentage_to_log = 0
        for i, namedTuple in enumerate(df.itertuples()):
            percentage_generated = int(100*i/column_length)
            if percentage_generated == percentage_to_log:
                self.log.debug(f"Processed {percentage_to_log}% of {self.generatedColumnName}")
                percentage_to_log += 5

            key = namedTuple.Index
            if self.cache is not None:
                value = self.cache.get(key)
                if value is None:
                    value = self._generate_value(namedTuple)
                    self.cache.set(key, value)
                else:
                    cache_hits += 1
            else:
                value = self._generate_value(namedTuple)
            values.append(value)
        if self.cache is not None:
            self.log.info(f"Cached column generation resulted in {cache_hits}/{column_length} cache hits")
        return values

    def __getstate__(self):
        if not self.persistCache:
            d = self.__dict__.copy()
            d["cache"] = None
            return d
        return self.__dict__

    @abstractmethod
    def _generate_value(self, named_tuple) -> Any:
        pass



import logging
from abc import ABC, abstractmethod
from copy import copy
from typing import List, Optional

import numpy as np
import pandas as pd

from sensai.util import mark_used

log = logging.getLogger(__name__)


class DataFrameColumnChangeTracker:
    """
    A simple class for keeping track of changes in columns between an initial data frame and some other data frame
    (usually the result of some transformations performed on the initial one).

    Example:

    >>> from sensai.util.pandas import DataFrameColumnChangeTracker
    >>> import pandas as pd

    >>> df = pd.DataFrame({"bar": [1, 2]})
    >>> columnChangeTracker = DataFrameColumnChangeTracker(df)
    >>> df["foo"] = [4, 5]
    >>> columnChangeTracker.track_change(df)
    >>> columnChangeTracker.get_removed_columns()
    set()
    >>> columnChangeTracker.get_added_columns()
    {'foo'}
    """
    def __init__(self, initial_df: pd.DataFrame):
        self.initialColumns = copy(initial_df.columns)
        self.final_columns = None

    def track_change(self, changed_df: pd.DataFrame):
        self.final_columns = copy(changed_df.columns)

    def get_removed_columns(self):
        self.assert_change_was_tracked()
        return set(self.initialColumns).difference(self.final_columns)

    def get_added_columns(self):
        """
        Returns the columns in the last entry of the history that were not present the first one
        """
        self.assert_change_was_tracked()
        return set(self.final_columns).difference(self.initialColumns)

    def column_change_string(self):
        """
        Returns a string representation of the change
        """
        self.assert_change_was_tracked()
        if list(self.initialColumns) == list(self.final_columns):
            return "none"
        removed_cols, added_cols = self.get_removed_columns(), self.get_added_columns()
        if removed_cols == added_cols == set():
            return f"reordered {list(self.final_columns)}"

        return f"added={list(added_cols)}, removed={list(removed_cols)}"

    def assert_change_was_tracked(self):
        if self.final_columns is None:
            raise Exception(f"No change was tracked yet. "
                            f"Did you forget to call trackChange on the resulting data frame?")


def extract_array(df: pd.DataFrame, dtype=None):
    """
    Extracts array from data frame. It is expected that each row corresponds to a data point and
    each column corresponds to a "channel". Moreover, all entries are expected to be arrays of the same shape
    (or scalars or sequences of the same length). We will refer to that shape as tensorShape.

    The output will be of shape `(N_rows, N_columns, *tensorShape)`. Thus, `N_rows` can be interpreted as dataset length
    (or batch size, if a single batch is passed) and N_columns can be interpreted as number of channels.
    Empty dimensions will be stripped, thus if the data frame has only one column, the array will have shape
    `(N_rows, *tensorShape)`.
    E.g. an image with three channels could equally be passed as data frame of the type


    +------------------+------------------+------------------+
    | R                | G                | B                |
    +==================+==================+==================+
    | channel          | channel          | channel          |
    +------------------+------------------+------------------+
    | channel          | channel          | channel          |
    +------------------+------------------+------------------+
    | ...              | ...              | ...              |
    +------------------+------------------+------------------+

    or as data frame of type

    +------------------+
    | image            |
    +==================+
    | RGB-array        |
    +------------------+
    | RGB-array        |
    +------------------+
    | ...              |
    +------------------+

    In both cases the returned array will have shape `(N_images, 3, width, height)`

    :param df: data frame where each entry is an array of shape tensorShape
    :param dtype: if not None, convert the array's data type to this type (string or numpy dtype)
    :return: array of shape `(N_rows, N_columns, *tensorShape)` with stripped empty dimensions
    """
    log.debug(f"Stacking tensors of shape {np.array(df.iloc[0, 0]).shape}")
    try:
        # This compact way of extracting the array causes dtypes to be modified,
        #    arr = np.stack(df.apply(np.stack, axis=1)).squeeze()
        # so we use this numpy-only alternative:
        arr = df.values
        if arr.shape[1] > 1:
            arr = np.stack([np.stack(arr[i]) for i in range(arr.shape[0])])
        else:
            arr = np.stack(arr[:, 0])
        # For the case where there is only one row, the old implementation above removed the first dimension,
        # so we do the same, even though it seems odd to do so (potential problem for batch size 1)
        # TODO: remove this behavior
        if arr.shape[0] == 1:
            arr = arr[0]
    except ValueError:
        raise ValueError(f"No array can be extracted from frame of length {len(df)} with columns {list(df.columns)}. "
                         f"Make sure that all entries have the same shape")
    if dtype is not None:
        arr = arr.astype(dtype, copy=False)
    return arr


def remove_duplicate_index_entries(df: pd.DataFrame):
    """
    Removes successive duplicate index entries by keeping only the first occurrence for every duplicate index element.

    :param df: the data frame, which is assumed to have a sorted index
    :return: the (modified) data frame with duplicate index entries removed
    """
    keep = [True]
    prev_item = df.index[0]
    for item in df.index[1:]:
        keep.append(item != prev_item)
        prev_item = item
    return df[keep]


def query_data_frame(df: pd.DataFrame, sql: str):
    """
    Queries the given data frame with the given condition specified in SQL syntax.

    NOTE: Requires duckdb to be installed.

    :param df: the data frame to query
    :param sql: an SQL query starting with the WHERE clause (excluding the 'where' keyword itself)
    :return: the filtered/transformed data frame
    """
    import duckdb

    NUM_TYPE_INFERENCE_ROWS = 100

    def is_supported_object_col(col_name: str):
        supported_type_set = set()
        contains_unsupported_types = False
        # check the first N values
        for value in df[col_name].iloc[:NUM_TYPE_INFERENCE_ROWS]:
            if isinstance(value, str):
                supported_type_set.add(str)
            elif value is None:
                pass
            else:
                contains_unsupported_types = True
        return not contains_unsupported_types and len(supported_type_set) == 1

    # determine which columns are object columns that are unsupported by duckdb and would raise errors
    # if they remained in the data frame that is queried
    added_index_col = "__sensai_resultset_index__"
    original_columns = df.columns
    object_columns = list(df.dtypes[df.dtypes == object].index)
    object_columns = [c for c in object_columns if not is_supported_object_col(c)]

    # add an artificial index which we will use to identify the rows for object column reconstruction
    df[added_index_col] = np.arange(len(df))

    try:
        # remove the object columns from the data frame but save them for subsequent reconstruction
        objects_df = df[object_columns + [added_index_col]]
        query_df = df.drop(columns=object_columns)
        mark_used(query_df)

        # apply query with reduced df
        result_df = duckdb.query(f"select * from query_df where {sql}").to_df()

        # restore object columns in result
        objects_df.set_index(added_index_col, drop=True, inplace=True)
        result_df.set_index(added_index_col, drop=True, inplace=True)
        result_objects_df = objects_df.loc[result_df.index]
        assert len(result_df) == len(result_objects_df)
        full_result_df = pd.concat([result_df, result_objects_df], axis=1)
        full_result_df = full_result_df[original_columns]

    finally:
        # clean up
        df.drop(columns=added_index_col, inplace=True)

    return full_result_df


class SeriesInterpolation(ABC):
    def interpolate(self, series: pd.Series, inplace: bool = False) -> Optional[pd.Series]:
        if not inplace:
            series = series.copy()
        self._interpolate_in_place(series)
        return series if not inplace else None

    @abstractmethod
    def _interpolate_in_place(self, series: pd.Series) -> None:
        pass

    def interpolate_all_with_combined_index(self, series_list: List[pd.Series]) -> List[pd.Series]:
        """
        Interpolates the given series using the combined index of all series.

        :param series_list: the list of series to interpolate
        :return: a list of corresponding interpolated series, each having the same index
        """
        # determine common index and
        index_elements = set()
        for series in series_list:
            index_elements.update(series.index)
        common_index = sorted(index_elements)

        # reindex, filling the gaps via interpolation
        interpolated_series_list = []
        for series in series_list:
            series = series.copy()
            series = series.reindex(common_index, method=None)
            self.interpolate(series, inplace=True)
            interpolated_series_list.append(series)

        return interpolated_series_list


class SeriesInterpolationLinearIndex(SeriesInterpolation):
    def __init__(self, ffill: bool = False, bfill: bool = False):
        """
        :param ffill: whether to fill any N/A values at the end of the series with the last valid observation
        :param bfill: whether to fill any N/A values at the start of the series with the first valid observation
        """
        self.ffill = ffill
        self.bfill = bfill

    def _interpolate_in_place(self, series: pd.Series) -> None:
        series.interpolate(method="index", inplace=True)
        if self.ffill:
            series.interpolate(method="ffill", limit_direction="forward")
        if self.bfill:
            series.interpolate(method="bfill", limit_direction="backward")


class SeriesInterpolationRepeatPreceding(SeriesInterpolation):
    def __init__(self, bfill: bool = False):
        """
        :param bfill: whether to fill any N/A values at the start of the series with the first valid observation
        """
        self.bfill = bfill

    def _interpolate_in_place(self, series: pd.Series) -> None:
        series.interpolate(method="pad", limit_direction="forward", inplace=True)
        if self.bfill:
            series.interpolate(method="bfill", limit_direction="backward")


def average_series(series_list: List[pd.Series], interpolation: SeriesInterpolation) -> pd.Series:
    interpolated_series_list = interpolation.interpolate_all_with_combined_index(series_list)
    return sum(interpolated_series_list) / len(interpolated_series_list)  # type: ignore



from abc import ABC, abstractmethod
from typing import Dict, Any, Optional, Generic, TypeVar, List

from matplotlib import pyplot as plt

from ..util import count_none
from ..util.deprecation import deprecated
from ..vector_model import VectorModelBase


class TrackingContext(ABC):
    def __init__(self, name: str, experiment: Optional["TrackedExperiment"]):
        # NOTE: `experiment` is optional only because of DummyTrackingContext
        self.name = name
        self._experiment = experiment
        self._isRunning = False

    @staticmethod
    def from_optional_experiment(experiment: Optional["TrackedExperiment"], model: Optional[VectorModelBase] = None,
            name: Optional[str] = None, description: str = ""):
        if experiment is None:
            return DummyTrackingContext(name)
        else:
            if count_none(name, model) != 1:
                raise ValueError("Must provide exactly one of {model, name}")
            if model is not None:
                return experiment.begin_context_for_model(model)
            else:
                return experiment.begin_context(name, description)

    def is_enabled(self):
        """
        :return: True if tracking is enabled, i.e. whether results can be saved via this context
        """
        return True

    @abstractmethod
    def _track_metrics(self, metrics: Dict[str, float]):
        pass

    def track_metrics(self, metrics: Dict[str, float], predicted_var_name: Optional[str] = None):
        """
        :param metrics: the metrics to be logged
        :param predicted_var_name: the name of the predicted variable for the case where there is more than one. If it is provided,
            the variable name will be prepended to every metric name.
        """
        if predicted_var_name is not None:
            metrics = {f"{predicted_var_name}_{k}": v for k, v in metrics.items()}
        self._track_metrics(metrics)

    @abstractmethod
    def track_figure(self, name: str, fig: plt.Figure):
        """
        :param name: the name of the figure (not a filename, should not include file extension)
        :param fig: the figure
        """
        pass

    @abstractmethod
    def track_text(self, name: str, content: str):
        """
        :param name: the name of the text (not a filename, should not include file extension)
        :param content: the content (arbitrarily long text, e.g. a log)
        """
        pass

    def __enter__(self):
        self._isRunning = True
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.end()

    @abstractmethod
    def _end(self):
        pass

    def end(self):
        # first end the context in the experiment (which may add final stuff)
        if self._isRunning:
            if self._experiment is not None:
                self._experiment.end_context(self)
            self._isRunning = False
        # then end the context for good
        self._end()


class DummyTrackingContext(TrackingContext):
    """
    A dummy tracking context which performs no actual tracking.
    It is useful to avoid having to write conditional tracking code for the case where there isn't a tracked experiment.
    """
    def __init__(self, name):
        super().__init__(name, None)

    def is_enabled(self):
        return False

    def _track_metrics(self, metrics: Dict[str, float]):
        pass

    def track_figure(self, name: str, fig: plt.Figure):
        pass

    def track_text(self, name: str, content: str):
        pass

    def _end(self):
        pass


TContext = TypeVar("TContext", bound=TrackingContext)


class TrackedExperiment(Generic[TContext], ABC):
    def __init__(self, context_prefix: str = "", additional_logging_values_dict=None):
        """
        Base class for tracking
        :param additional_logging_values_dict: additional values to be logged for each run
        """
        # TODO additional_logging_values_dict probably needs to be removed
        self.instancePrefix = context_prefix
        self.additionalLoggingValuesDict = additional_logging_values_dict
        self._contexts: List[TContext] = []

    @deprecated("Use a tracking context instead")
    def track_values(self, values_dict: Dict[str, Any], add_values_dict: Dict[str, Any] = None):
        values_dict = dict(values_dict)
        if add_values_dict is not None:
            values_dict.update(add_values_dict)
        if self.additionalLoggingValuesDict is not None:
            values_dict.update(self.additionalLoggingValuesDict)
        self._track_values(values_dict)

    @abstractmethod
    def _track_values(self, values_dict: Dict[str, Any]):
        pass

    @abstractmethod
    def _create_tracking_context(self, name: str, description: str) -> TContext:
        pass

    def begin_context(self, name: str, description: str = "") -> TContext:
        """
        Begins a context in which actual information will be tracked.
        The returned object is a context manager, which can be used in a with-statement.

        :param name: the name of the context (e.g. model name)
        :param description: a description (e.g. full model parameters/specification)
        :return: the context, which can subsequently be used to track information
        """
        instance = self._create_tracking_context(self.instancePrefix + name, description)
        self._contexts.append(instance)
        return instance

    def begin_context_for_model(self, model: VectorModelBase):
        """
        Begins a tracking context for the case where we want to track information about a model (wrapper around `begin_context` for convenience).
        The model name is used as the context name, and the model's string representation is used as the description.
        The returned object is a context manager, which can be used in a with-statement.

        :param model: the model
        :return: the context, which can subsequently be used to track information
        """
        return self.begin_context(model.get_name(), model.pprints())

    def end_context(self, instance: TContext):
        running_instance = self._contexts[-1]
        if instance != running_instance:
            raise ValueError(f"Passed instance ({instance}) is not the currently running instance ({running_instance})")
        self._contexts.pop()

    def __del__(self):
        # make sure all contexts that are still running are eventually closed
        for c in reversed(self._contexts):
            c.end()


class TrackingMixin(ABC):
    _objectId2trackedExperiment = {}

    def set_tracked_experiment(self, tracked_experiment: Optional[TrackedExperiment]):
        self._objectId2trackedExperiment[id(self)] = tracked_experiment

    def unset_tracked_experiment(self):
        self.set_tracked_experiment(None)

    @property
    def tracked_experiment(self) -> Optional[TrackedExperiment]:
        return self._objectId2trackedExperiment.get(id(self))

    def begin_optional_tracking_context_for_model(self, model: VectorModelBase, track: bool = True) -> TrackingContext:
        """
        Begins a tracking context for the given model; the returned object is a context manager and therefore method should
        preferably be used in a `with` statement.
        This method can be called regardless of whether there actually is a tracked experiment (hence the term 'optional').
        If there is no tracked experiment, calling methods on the returned object has no effect.
        Furthermore, tracking can be disabled by passing `track=False` even if a tracked experiment is present.

        :param model: the model for which to begin tracking
        :paraqm track: whether tracking shall be enabled; if False, force use of a dummy context which performs no actual tracking even
            if a tracked experiment is present
        :return: a context manager that can be used to track results for the given model
        """
        return TrackingContext.from_optional_experiment(self.tracked_experiment if track else None, model=model)



"""
Utility classes and functions for JavaScript code generation
"""
import json
from abc import abstractmethod, ABC
from typing import Union, Any

import numpy as np

from .string import list_string


PythonType = Union[str, int, bool, float]


class JsCode(ABC):
    def __str__(self):
        return self.get_js_code()

    @abstractmethod
    def get_js_code(self):
        pass


class JsCodeLiteral(JsCode):
    def __init__(self, js_code: str):
        self.js_code = js_code

    def get_js_code(self):
        return self.js_code


class JsValue(JsCode, ABC):
    @classmethod
    def from_python(cls, value: PythonType):
        t = type(value)
        if t == str:
            return cls.string_value(value)
        elif t == int:
            return cls.int_value(value)
        elif value is None:
            return cls.undefined()
        elif t == bool:
            return cls.bool_value(value)
        elif t in (float, np.float64, np.float):
            return cls.float_value(value)
        else:
            raise ValueError(f"Unsupported value of type {type(value)}: {value}")

    @classmethod
    def from_value(cls, value: Union["JsValue", PythonType]):
        if isinstance(value, JsValue):
            return value
        else:
            return cls.from_python(value)

    def is_undefined(self):
        return self.get_js_code() == "undefined"

    @staticmethod
    def string_value(s: str):
        s = s.replace('"', r'\"')
        return JsValueLiteral(f'"{s}"')

    @staticmethod
    def int_value(value: int):
        return JsValueLiteral(str(int(value)))

    @staticmethod
    def float_value(value: Union[float, int]):
        return JsValueLiteral(str(float(value)))

    @staticmethod
    def bool_value(value: bool):
        b = bool(value)
        return JsValueLiteral("true" if b else "false")

    @staticmethod
    def undefined():
        return JsValueLiteral("undefined")

    @staticmethod
    def null():
        return JsValueLiteral("null")


class JsValueLiteral(JsValue):
    def __init__(self, js_code: str):
        self.js_code = js_code

    def get_js_code(self):
        return self.js_code


def js_value(value: Union[JsValue, PythonType]) -> JsValue:
    return JsValue.from_value(value)


def js_arg_list(*args: Union[JsValue, PythonType], drop_trailing_undefined=True) -> JsCode:
    """
    :param args: arguments that are either JsValue instances or (supported) Python values
    :param drop_trailing_undefined: whether to drop trailing arguments that are undefined/None
    :return: the JsCode
    """
    args = [js_value(a) for a in args]
    last_index_to_include = len(args) - 1
    if drop_trailing_undefined:
        while last_index_to_include >= 0 and args[last_index_to_include].is_undefined():
            last_index_to_include -= 1
    args = args[:last_index_to_include+1]
    return JsCodeLiteral(", ".join(map(str, args)))


class JsObject(JsValue):
    def __init__(self):
        self.data = {}

    def add(self, key: str, value: Union[JsValue, PythonType]):
        self.data[key] = js_value(value)

    def add_string(self, key: str, value: str):
        self.data[key] = JsValue.string_value(value)

    def add_code_literal(self, key: str, value: str):
        self.data[key] = value

    def add_float(self, key: str, value: Union[float, int]):
        self.data[key] = JsValue.float_value(value)

    def add_json(self, key: str, value: Any):
        """
        :param key: key within the object
        :param value: any Python object which can be converted to JSON
        """
        self.add_code_literal(key, json.dumps(value))

    def get_js_code(self):
        return "{" + ", ".join(f'"{k}": {v}' for k, v in self.data.items()) + "}"

    def __len__(self):
        return len(self.data)


class JsClassInstance(JsValueLiteral):
    def __init__(self, class_name, *args: Union[JsValue, PythonType]):
        arg_list = js_arg_list(*args, drop_trailing_undefined=False)
        super().__init__(f"new {class_name}({arg_list})")


class JsList(JsValueLiteral):
    def __init__(self, *values: Union[JsValue, PythonType]):
        super().__init__(list_string([js_value(x) for x in values]))



import copy
import functools
import logging
from abc import ABC, abstractmethod
from typing import Tuple, Any, Generator, Generic, TypeVar, List, Union, Sequence, Optional

import numpy as np

from .eval_stats.eval_stats_base import PredictionEvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats, ClassificationEvalStatsCollection
from .eval_stats.eval_stats_regression import RegressionEvalStats, RegressionEvalStatsCollection
from .evaluator import VectorRegressionModelEvaluationData, VectorClassificationModelEvaluationData, \
    VectorModelEvaluationData, VectorClassificationModelEvaluator, VectorRegressionModelEvaluator, \
    MetricsDictProvider, VectorModelEvaluator, ClassificationEvaluatorParams, \
    RegressionEvaluatorParams, MetricsDictProviderFromFunction
from ..data import InputOutputData, DataSplitterFractional
from ..tracking.tracking_base import TrackingContext
from ..util.typing import PandasNamedTuple
from ..vector_model import VectorClassificationModel, VectorRegressionModel, VectorModel

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=PredictionEvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)
TEvalData = TypeVar("TEvalData", bound=VectorModelEvaluationData)


class VectorModelCrossValidationData(ABC, Generic[TModel, TEvalData, TEvalStats, TEvalStatsCollection]):
    def __init__(self, trained_models: Optional[List[TModel]], eval_data_list: List[TEvalData], predicted_var_names: List[str],
            test_indices_list=None):
        self.predicted_var_names = predicted_var_names
        self.trained_models = trained_models
        self.eval_data_list = eval_data_list
        self.test_indices_list = test_indices_list

    @property
    def model_name(self):
        return self.eval_data_list[0].model_name

    @abstractmethod
    def _create_eval_stats_collection(self, l: List[TEvalStats]) -> TEvalStatsCollection:
        pass

    def get_eval_stats_collection(self, predicted_var_name=None) -> TEvalStatsCollection:
        if predicted_var_name is None:
            if len(self.predicted_var_names) != 1:
                raise Exception(f"Must provide name of predicted variable name, as multiple variables were predicted: "
                                f"{self.predicted_var_names}")
            else:
                predicted_var_name = self.predicted_var_names[0]
        eval_stats_list = [evalData.get_eval_stats(predicted_var_name) for evalData in self.eval_data_list]
        return self._create_eval_stats_collection(eval_stats_list)

    def iter_input_output_ground_truth_tuples(self, predicted_var_name=None) -> Generator[Tuple[PandasNamedTuple, Any, Any], None, None]:
        for evalData in self.eval_data_list:
            eval_stats = evalData.get_eval_stats(predicted_var_name)
            for i, namedTuple in enumerate(evalData.input_data.itertuples()):
                yield namedTuple, eval_stats.y_predicted[i], eval_stats.y_true[i]

    def track_metrics(self, tracking_context: TrackingContext):
        is_multivar = len(self.predicted_var_names) > 1
        for predicted_var_name in self.predicted_var_names:
            eval_stats_collection = self.get_eval_stats_collection(predicted_var_name=predicted_var_name)
            metrics_dict = eval_stats_collection.agg_metrics_dict()
            tracking_context.track_metrics(metrics_dict, predicted_var_name=predicted_var_name if is_multivar else None)


TCrossValData = TypeVar("TCrossValData", bound=VectorModelCrossValidationData)


class CrossValidationSplitter(ABC):
    """
    Defines a mechanism with which to generate data splits for cross-validation
    """
    @abstractmethod
    def create_folds(self, data: InputOutputData, num_folds: int) -> List[Tuple[Sequence[int], Sequence[int]]]:
        """
        :param data: the data from which to obtain the folds
        :param num_folds: the number of splits/folds
        :return: a list containing numFolds tuples (t, e) where t and e are sequences of data point indices to use for training
            and evaluation respectively
        """
        pass


class CrossValidationSplitterDefault(CrossValidationSplitter):
    def __init__(self, shuffle=True, random_seed=42):
        self.shuffle = shuffle
        self.randomSeed = random_seed

    def create_folds(self, data: InputOutputData, num_splits: int) -> List[Tuple[Sequence[int], Sequence[int]]]:
        num_data_points = len(data)
        num_test_points = num_data_points // num_splits
        if self.shuffle:
            indices = np.random.RandomState(self.randomSeed).permutation(num_data_points)
        else:
            indices = list(range(num_data_points))
        result = []
        for i in range(num_splits):
            test_start_idx = i * num_test_points
            test_end_idx = test_start_idx + num_test_points
            test_indices = indices[test_start_idx:test_end_idx]
            train_indices = np.concatenate((indices[:test_start_idx], indices[test_end_idx:]))
            result.append((train_indices, test_indices))
        return result


class CrossValidationSplitterNested(CrossValidationSplitter):
    """
    A data splitter for nested cross-validation (which is useful, in particular, for time series prediction problems)
    """
    def __init__(self, test_fraction: float):
        self.test_fraction = test_fraction

    def create_folds(self, data: InputOutputData, num_folds: int) -> List[Tuple[Sequence[int], Sequence[int]]]:
        fractional_splitter = DataSplitterFractional(1-self.test_fraction, shuffle=False)
        result = []
        for i in range(num_folds):
            indices, (a, b) = fractional_splitter.split_with_indices(data)
            result.append(indices)
            data = a
        return result


class VectorModelCrossValidatorParams:
    def __init__(self,
            folds: int = 5,
            splitter: CrossValidationSplitter = None,
            return_trained_models=False,
            evaluator_params: Union[RegressionEvaluatorParams, ClassificationEvaluatorParams] = None,
            default_splitter_random_seed=42,
            default_splitter_shuffle=True):
        """
        :param folds: the number of folds
        :param splitter: the splitter to use in order to generate the folds; if None, use default split (using parameters for random seed
            and shuffling below)
        :param return_trained_models: whether to create a copy of the model for each fold and return each of the models
            (requires that models can be deep-copied); if False, the model that is passed to evalModel is fitted several times
        :param evaluator_params: the model evaluator parameters
        :param default_splitter_random_seed: [if splitter is None] the random seed to use for splits
        :param default_splitter_shuffle: [if splitter is None] whether to shuffle the data (using randomSeed) before creating the folds
        """
        self.folds = folds
        self.evaluatorParams = evaluator_params
        self.returnTrainedModels = return_trained_models
        if splitter is None:
            splitter = CrossValidationSplitterDefault(shuffle=default_splitter_shuffle, random_seed=default_splitter_random_seed)
        self.splitter = splitter


class VectorModelCrossValidator(MetricsDictProvider, Generic[TCrossValData], ABC):
    def __init__(self, data: InputOutputData, params: Union[VectorModelCrossValidatorParams]):
        """
        :param data: the data set
        :param params: parameters
        """
        self.params = params
        self.modelEvaluators: List[VectorModelEvaluator] = []
        for trainIndices, testIndices in self.params.splitter.create_folds(data, self.params.folds):
            self.modelEvaluators.append(self._create_model_evaluator(data.filter_indices(trainIndices), data.filter_indices(testIndices)))

    @staticmethod
    def for_model(model: VectorModel, data: InputOutputData, params: VectorModelCrossValidatorParams) \
            -> Union["VectorClassificationModelCrossValidator", "VectorRegressionModelCrossValidator"]:
        if model.is_regression_model():
            return VectorRegressionModelCrossValidator(data, params)
        else:
            return VectorClassificationModelCrossValidator(data, params)

    @abstractmethod
    def _create_model_evaluator(self, training_data: InputOutputData, test_data: InputOutputData) -> VectorModelEvaluator:
        pass

    @abstractmethod
    def _create_result_data(self, trained_models, eval_data_list, test_indices_list, predicted_var_names) -> TCrossValData:
        pass

    def eval_model(self, model: VectorModel, track: bool = True):
        """
        :param model: the model to evaluate
        :param track: whether tracking shall be enabled for the case where a tracked experiment is set on this object
        :return: cross-validation results
        """
        trained_models = [] if self.params.returnTrainedModels else None
        eval_data_list = []
        test_indices_list = []
        predicted_var_names = None
        with self.begin_optional_tracking_context_for_model(model, track=track) as tracking_context:
            for i, evaluator in enumerate(self.modelEvaluators, start=1):
                evaluator: VectorModelEvaluator
                log.info(f"Training and evaluating model with fold {i}/{len(self.modelEvaluators)} ...")
                model_to_fit: VectorModel = copy.deepcopy(model) if self.params.returnTrainedModels else model
                evaluator.fit_model(model_to_fit)
                eval_data = evaluator.eval_model(model_to_fit)
                if predicted_var_names is None:
                    predicted_var_names = eval_data.predicted_var_names
                if self.params.returnTrainedModels:
                    trained_models.append(model_to_fit)
                for predictedVarName in predicted_var_names:
                    log.info(f"Evaluation result for {predictedVarName}, fold {i}/{len(self.modelEvaluators)}: "
                             f"{eval_data.get_eval_stats(predicted_var_name=predictedVarName)}")
                eval_data_list.append(eval_data)
                test_indices_list.append(evaluator.test_data.outputs.index)
            crossval_data = self._create_result_data(trained_models, eval_data_list, test_indices_list, predicted_var_names)
            if tracking_context.is_enabled():
                crossval_data.track_metrics(tracking_context)
        return crossval_data

    def _compute_metrics(self, model: VectorModel, **kwargs):
        return self._compute_metrics_for_var_name(model, None)

    def _compute_metrics_for_var_name(self, model, predicted_var_name: Optional[str]):
        data = self.eval_model(model)
        return data.get_eval_stats_collection(predicted_var_name=predicted_var_name).agg_metrics_dict()

    def create_metrics_dict_provider(self, predicted_var_name: Optional[str]) -> MetricsDictProvider:
        """
        Creates a metrics dictionary provider, e.g. for use in hyperparameter optimisation

        :param predicted_var_name: the name of the predicted variable for which to obtain evaluation metrics; may be None only
            if the model outputs but a single predicted variable
        :return: a metrics dictionary provider instance for the given variable
        """
        return MetricsDictProviderFromFunction(functools.partial(self._compute_metrics_for_var_name, predictedVarName=predicted_var_name))


class VectorRegressionModelCrossValidationData(VectorModelCrossValidationData[VectorRegressionModel, VectorRegressionModelEvaluationData,
        RegressionEvalStats, RegressionEvalStatsCollection]):
    def _create_eval_stats_collection(self, l: List[RegressionEvalStats]) -> RegressionEvalStatsCollection:
        return RegressionEvalStatsCollection(l)


class VectorRegressionModelCrossValidator(VectorModelCrossValidator[VectorRegressionModelCrossValidationData]):
    def _create_model_evaluator(self, training_data: InputOutputData, test_data: InputOutputData) -> VectorRegressionModelEvaluator:
        evaluator_params = RegressionEvaluatorParams.from_dict_or_instance(self.params.evaluatorParams)
        return VectorRegressionModelEvaluator(training_data, test_data=test_data, params=evaluator_params)

    def _create_result_data(self, trained_models, eval_data_list, test_indices_list, predicted_var_names) \
            -> VectorRegressionModelCrossValidationData:
        return VectorRegressionModelCrossValidationData(trained_models, eval_data_list, predicted_var_names, test_indices_list)


class VectorClassificationModelCrossValidationData(VectorModelCrossValidationData[VectorClassificationModel,
        VectorClassificationModelEvaluationData, ClassificationEvalStats, ClassificationEvalStatsCollection]):
    def _create_eval_stats_collection(self, l: List[ClassificationEvalStats]) -> ClassificationEvalStatsCollection:
        return ClassificationEvalStatsCollection(l)


class VectorClassificationModelCrossValidator(VectorModelCrossValidator[VectorClassificationModelCrossValidationData]):
    def _create_model_evaluator(self, training_data: InputOutputData, test_data: InputOutputData):
        evaluator_params = ClassificationEvaluatorParams.from_dict_or_instance(self.params.evaluatorParams)
        return VectorClassificationModelEvaluator(training_data, test_data=test_data, params=evaluator_params)

    def _create_result_data(self, trained_models, eval_data_list, test_indices_list, predicted_var_names) \
            -> VectorClassificationModelCrossValidationData:
        return VectorClassificationModelCrossValidationData(trained_models, eval_data_list, predicted_var_names, test_indices_list)



import collections
from typing import Hashable, Dict, Optional

from .string import ToStringMixin


class RelativeFrequencyCounter(ToStringMixin):
    """
    Counts the absolute and relative frequency of an event
    """
    def __init__(self):
        self.num_total = 0
        self.num_relevant = 0

    def count(self, is_relevant_event) -> None:
        """
        Adds to the count.
        The nominator is incremented only if we are counting a relevant event.
        The denominator is always incremented.

        :param is_relevant_event: whether we are counting a relevant event
        """
        self.num_total += 1
        if is_relevant_event:
            self.num_relevant += 1

    def _tostring_object_info(self):
        info = f"{self.num_relevant}/{self.num_total}"
        if self.num_total > 0:
            info += f", {100 * self.num_relevant / self.num_total:.2f}%"
        return info

    def add(self, relative_frequency_counter: "RelativeFrequencyCounter") -> None:
        """
        Adds the counts of the given counter to this object

        :param relative_frequency_counter: the counter whose data to add
        """
        self.num_total += relative_frequency_counter.num_total
        self.num_relevant += relative_frequency_counter.num_relevant

    def get_relative_frequency(self) -> Optional[float]:
        """
        :return: the relative frequency (between 0 and 1) or None if nothing was counted (0 events considered)
        """
        if self.num_total == 0:
            return None
        return self.num_relevant / self.num_total


class DistributionCounter(ToStringMixin):
    """
    Supports the counting of the frequencies with which (mutually exclusive) events occur
    """
    def __init__(self):
        self.counts = collections.defaultdict(self._zero)
        self.total_count = 0

    @staticmethod
    def _zero():
        return 0

    def count(self, event: Hashable) -> None:
        """
        Increments the count of the given event

        :param event: the event/key whose count to increment, which must be hashable
        """
        self.total_count += 1
        self.counts[event] += 1

    def get_distribution(self) -> Dict[Hashable, float]:
        """
        :return: a dictionary mapping events (as previously passed to count) to their relative frequencies
        """
        return {k: v/self.total_count for k, v in self.counts.items()}

    def _tostring_object_info(self):
        return ", ".join([f"{str(k)}: {v} ({v/self.total_count:.3f})" for k, v in self.counts.items()])


class WeightedMean(ToStringMixin):
    """
    Computes a weighted mean of values
    """
    def __init__(self):
        self.weighted_value_sum = 0
        self.weight_sum = 0

    def _tostring_object_info(self) -> str:
        return f"{self.weighted_value_sum / self.weight_sum}"

    def add(self, value, weight=1) -> None:
        """
        Adds the given value with the given weight to the calculation

        :param value: the value
        :param weight: the weight with which to consider the value
        """
        self.weighted_value_sum += value * weight
        self.weight_sum += weight

    def get_weighted_mean(self):
        """
        :return: the weighted mean of all values that have been added
        """
        return self.weighted_value_sum / self.weight_sum



"""
This module contains base classes for models that input and output tensors, for examples CNNs.
The fitting and predictions will still be performed on data frames, like in VectorModel,
but now it will be expected that all entries of the input data frame passed to the model are tensors of the same shape.
Lists of scalars of the same lengths are also accepted. The same is expected of the ground truth data frames.
Everything will work as well if the entries are just scalars but in this case it is recommended to use
VectorModel instead.

If we denote the shapes of entries in the dfs as inputTensorShape and outputTensorShape,
the model will be fit on input tensors of shape (N_rows, N_inputColumns, inputTensorShape) and output tensors of
shape (N_rows, N_outputColumns, outputTensorShape), where empty dimensions (e.g. for one-column data frames)
will be stripped.
"""


import logging
from abc import ABC, abstractmethod
from typing import Optional, Tuple

import numpy as np
import pandas as pd

from .util.pandas import extract_array
from .vector_model import VectorRegressionModel, VectorClassificationModel, VectorModel

log = logging.getLogger(__name__)
# we set the default level to debug because it is often interesting for the user to receive
# debug information about shapes as data frames get converted to arrays
log.setLevel(logging.DEBUG)


class InvalidShapeError(Exception):
    pass


def _get_datapoint_shape(df: pd.DataFrame):
    first_row_df = df.iloc[:1]
    # Note that the empty first dimension with N_Datapoints=1 is stripped by extractArray
    return extract_array(first_row_df).shape


def _check_df_shape(df: pd.DataFrame, desired_shape: tuple):
    datapoint_shape = _get_datapoint_shape(df)
    if datapoint_shape != desired_shape:
        raise InvalidShapeError(f"Wrong input shape for data point. Expected {desired_shape} but got {datapoint_shape}")


# This is implemented as a mixin because there can be no functional common class for all tensor models.
# The reason is that actual implementations need to inherit from Vector-Regression/Classification-Model
# (or duplicate a lot of code) and thus it is not possible to inherit from something like TensorModel(VectorModel)
# without getting into a mess.
class TensorModel(ABC):
    def __init__(self):
        self._modelInputShape = None
        self._modelOutputShape = None

    @abstractmethod
    def _fit_to_array(self, x: np.ndarray, y: np.ndarray):
        pass

    @abstractmethod
    def _predict_array(self, x: np.ndarray) -> np.ndarray:
        """
        The result should be of shape `(N_DataPoints, *predictedTensorShape)` if a single column is predicted
        or of shape `(N_DataPoints, N_Columns, *predictedTensorShape)` if multiple columns are predicted
        (e.g. for multiple regression targets). Note that in both cases, the number of predicted columns
        should coincide with corresponding number in the ground truth data frame the model was fitted on

        :param x: a tensor of shape `(N_DataPoints, *inputTensorShape)`
        """
        pass

    def _predict_df_through_array(self, x: pd.DataFrame, output_columns: list) -> pd.DataFrame:
        """
        To be used within _predict in implementations of this class. Performs predictions by
        transforming X into an array, computing the predicted array from it and turning the result into a
        predictions data frame.

        :param x: input data frame (of same type as for _fitTensorModel)
        :param output_columns: columns of the outputDF, typically the result of calling `getPredictedVariableNames()` in
            an implementation
        :return:
        """
        y = self._predict_array(extract_array(x))
        if not len(y) == len(x):
            raise InvalidShapeError(f"Number of data points (lengths) of input data frame and predictions must agree. "
                                    f"Expected {len(x)} but got {len(y)}")

        result = pd.DataFrame(index=x.index)
        n_columns = len(output_columns)
        if n_columns == 1:
            result[output_columns[0]] = list(y)
        else:

            if not n_columns == y.shape[1]:
                raise InvalidShapeError(f"Wrong shape of predictions array for a data frame with {n_columns} columns ({output_columns}). "
                                        f"Expected shape ({len(x)}, {n_columns}, ...) but got: {y.shape}")
            for i, col in enumerate(output_columns):
                result[col] = list(y[:, i])
        return result

    def _fit_tensor_model(self, x: pd.DataFrame, y: pd.DataFrame):
        """
        To be used within _fit in implementations of this class
        """
        log.debug(f"Stacking input tensors from columns {x.columns} and from all rows to a single array. "
                  f"Note that all tensors need to have the same shape")
        x = extract_array(x)
        y = extract_array(y)
        self._modelInputShape = x[0].shape
        self._modelOutputShape = y[0].shape
        log.debug(f"Fitting on {len(x)} datapoints of shape {self._modelInputShape}. "
                  f"The ground truth are tensors of shape {self._modelOutputShape}")
        self._fit_to_array(x, y)

    def get_model_input_shape(self) -> Optional[Tuple]:
        return self._modelInputShape

    def get_model_output_shape(self):
        return self._modelInputShape


class TensorToScalarRegressionModel(VectorRegressionModel, TensorModel, ABC):
    def __init__(self, check_input_shape=True, check_input_columns=True):
        """
        Base class for regression models that take tensors as input and output scalars. They can be evaluated
        in the same way as non-tensor regression models

        :param check_input_shape: Whether to check if during predict input tensors have the same shape as during fit.
            For certain applications, e.g. using CNNs on larger inputs than the training set, this has
            to be disabled
        :param check_input_columns: Whether to check if input columns at predict time coincide with those at fit time
        """
        VectorRegressionModel.__init__(self, check_input_columns=check_input_columns)
        TensorModel.__init__(self)
        self.check_input_shape = check_input_shape

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        self._warn_sample_weights_unsupported(False, weights)
        self._fit_tensor_model(x, y)

    def _predict(self, x: pd.DataFrame) -> pd.DataFrame:
        return self._predict_df_through_array(x, self.get_predicted_variable_names())

    def predict(self, x: pd.DataFrame) -> pd.DataFrame:
        if self.check_input_shape:
            _check_df_shape(x, self.get_model_input_shape())
        return super().predict(x)


class TensorToScalarClassificationModel(VectorClassificationModel, TensorModel, ABC):
    def __init__(self, check_input_shape=True, check_input_columns=True):
        """
        Base class for classification models that take tensors as input and output scalars. They can be evaluated
        in the same way as non-tensor classification models

        :param check_input_shape: Whether to check if during predict input tensors have the same shape as during fit.
            For certain applications, e.g. using CNNs on larger inputs than the training set, this has
            to be disabled
        :param check_input_columns: Whether to check if input columns at predict time coincide with those at fit time
        """
        VectorClassificationModel.__init__(self, check_input_columns=check_input_columns)
        TensorModel.__init__(self)
        self.checkInputShape = check_input_shape

    def _predict_class_probabilities(self, x: pd.DataFrame) -> pd.DataFrame:
        return self._predict_df_through_array(x, self.get_class_labels())

    def _fit_classifier(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        self._warn_sample_weights_unsupported(False, weights)
        self._fit_tensor_model(x, y)

    def predict(self, x: pd.DataFrame) -> pd.DataFrame:
        if self.checkInputShape:
            _check_df_shape(x, self.get_model_input_shape())
        return super().predict(x)

    # just renaming the abstract method to implement
    def _predict_array(self, x: np.ndarray) -> np.ndarray:
        return self._predict_probabilities_array(x)

    @abstractmethod
    def _predict_probabilities_array(self, x: np.ndarray) -> np.ndarray:
        """
        If you are implementing a probabilistic classifier, this method should return a tensor with probabilities
        of shape `(N_DataPoints, N_Labels)`. It is assumed that labels are lexicographically sorted and the order
        of predictions in the output array should respect this.

        The default implementation of _predict will then use the output of this method and convert it to predicted labels (via argmax).

        In case you want to predict labels only or have a more efficient implementation of predicting labels than
        using argmax, your will have to override _predict in your implementation. In the former case of a
        non-probabilistic classifier, the implementation of this method should raise an exception, like the one below.
        """
        raise NotImplementedError(f"Model {self.__class__.__name__} does not support prediction of probabilities")


# Note: for tensor to tensor models the output shape is not trivial. There will be dedicated evaluators
# and metrics for them. Examples for such models are auto-encoders, models performing semantic segregation,
# models for super-resolution and so on
class TensorToTensorRegressionModel(VectorRegressionModel, TensorModel, ABC):
    def __init__(self, check_input_shape=True, check_output_shape=True, check_input_columns=True):
        """
        Base class for regression models that output tensors. Multiple targets can be used by putting
        them into separate columns. In that case it is required that all target tensors have the same shape.

        :param check_input_shape: Whether to check if during predict tensors have the same shape as during fit.
            For certain applications, e.g. using CNNs on larger inputs than the training set, this has
            to be disabled
        :param check_output_shape: Whether to check if predictions have the same shape as ground truth data during fit.
            For certain applications, e.g. using CNNs on larger inputs than the training set, this has
            to be disabled
        :param check_input_columns: Whether to check if input columns at predict time coincide with those at fit time
        """
        VectorRegressionModel.__init__(self, check_input_columns=check_input_columns)
        TensorModel.__init__(self)
        self.checkInputShape = check_input_shape
        self.checkOutputShape = check_output_shape

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        self._warn_sample_weights_unsupported(False, weights)
        self._fit_tensor_model(x, y)

    def _predict(self, x: pd.DataFrame) -> pd.DataFrame:
        return self._predict_df_through_array(x, self.get_predicted_variable_names())

    def predict(self, x: pd.DataFrame) -> pd.DataFrame:
        if not self.is_fitted():
            raise Exception(f"Calling predict with unfitted model. "
                            f"This might lead to errors down the line, especially if input/output checks are enabled")
        if self.checkInputShape:
            _check_df_shape(x, self.get_model_input_shape())
        y = super().predict(x)
        if self.checkOutputShape:
            _check_df_shape(y, self.get_model_output_shape())
        return y


class TensorToTensorClassificationModel(VectorModel, TensorModel, ABC):
    def __init__(self, check_input_shape=True, check_output_shape=True, check_input_columns=True):
        """
        Base class for classification models that output tensors, e.g. for semantic segregation. The models
        can be fit on a ground truth data frame with a single column. The entries in this column should be
        binary tensors with one-hot-encoded labels, i.e. of shape `(*predictionShape, numLabels)`

        :param check_input_shape: Whether to check if during predict tensors have the same shape as during fit.
            For certain applications, e.g. using CNNs on larger inputs than the training set, this has
            to be disabled
        :param check_output_shape: Whether to check if predictions have the same shape as ground truth data during fit.
            For certain applications, e.g. using CNNs on larger inputs than the training set, this has
            to be disabled
        :param check_input_columns: Whether to check if input columns at predict time coincide with those at fit time
        """
        VectorModel.__init__(self, check_input_columns=check_input_columns)
        TensorModel.__init__(self)
        self.check_input_shape = check_input_shape
        self.check_output_shape = check_output_shape
        self._numPredictedClasses: Optional[int] = None

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        self._warn_sample_weights_unsupported(False, weights)
        self._fit_tensor_model(x, y)

    def is_regression_model(self) -> bool:
        return False

    def get_num_predicted_classes(self):
        return self._numPredictedClasses

    def fit(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None, fit_preprocessors=True, fit_model=True):
        """

        :param x: data frame containing input tensors on which to train
        :param y: ground truth has to be an array containing only zeroes and ones (one-hot-encoded labels) of the shape
            `(*prediction_shape, numLabels)`
        :param weights: data point weights (must be None; not supported by this model!)

        :param fit_preprocessors: whether the model's preprocessors (feature generators and data frame transformers) shall be fitted
        :param fit_model: whether the model itself shall be fitted
        """
        if len(y.columns) != 1:
            raise ValueError(f"{self.__class__.__name__} requires exactly one output "
                             f"column with tensors containing one-hot-encoded labels")

        # checking if Y is a binary array of one hot encoded labels
        df_y_to_check = extract_array(y.iloc[:5])
        if not np.array_equal(df_y_to_check, df_y_to_check.astype(bool)):
            raise Exception(f"Ground truth data points have to be binary arrays of one-hot-encoded labels "
                            f"of shape (*prediction_shape, numLabels). Did you forget to one-hot-encode your labels "
                            f"before training?")
        # df_y_to_check has shape (N_datapoints=5, *prediction_shape, N_labels)
        prediction_shape = df_y_to_check.shape[1:-1]
        if len(prediction_shape) == 0:
            raise InvalidShapeError(f"Ground truth data points have to be binary arrays of one-hot-encoded labels "
                                    f"of shape (*prediction_shape, numLabels). However, received array of trivial "
                                    f"prediction_shape. If the predictions are scalars, a TensorToScalarClassificationModel "
                                    f"should be used instead of {self.__class__.__name__}")
        self._numPredictedClasses = df_y_to_check.shape[-1]
        super().fit(x, y, weights=weights, fit_preprocessors=fit_preprocessors, fit_model=True)

    def get_model_output_shape(self):
        # The ground truth contains one-hot-encoded labels in the last dimension
        # The model output predicts the labels as ints, without one-hot-encoding
        one_hot_encoded_output_shape = super().get_model_output_shape()
        if one_hot_encoded_output_shape is None:
            return None
        return one_hot_encoded_output_shape[:-1]

    def convert_class_probabilities_to_predictions(self, df: pd.DataFrame):
        """
        Converts from a result returned by predictClassProbabilities to a result as return by predict.

        :param df: the output data frame from predictClassProbabilities
        :return: an output data frame as it would be returned by predict
        """
        df = df.copy()
        col_name = self.get_predicted_variable_names()[0]
        df[col_name] = df[col_name].apply(lambda probas_array: probas_array.argmax(axis=-1))
        return df

    def predict_class_probabilities(self, x: pd.DataFrame) -> pd.DataFrame:
        """
        :param x: the input data
        :return: a data frame with a single column containing arrays of shape `(*tensorShape, numLabels)`.
             Raises an exception if the classifier cannot predict probabilities.
        """
        x = self._compute_model_inputs(x)
        if self.check_input_shape:
            _check_df_shape(x, self.get_model_input_shape())
        result = self._predict_class_probabilities(x)
        self._check_prediction(result)
        return result

    def _check_prediction(self, prediction_df: pd.DataFrame, max_rows_to_check=5):
        """
        Checks whether the column name is correctly, whether the shapes match the ground truth and whether the entries
        correspond to probabilities
        """
        if self.check_output_shape:
            _check_df_shape(prediction_df, self.get_model_output_shape())

        array_to_check = extract_array(prediction_df.iloc[:max_rows_to_check])

        if not np.all(0 <= array_to_check) or not np.all(array_to_check <= 1):
            log.warning(f"Probability arrays may not be correctly normalised, "
                        f"got probabilities outside the range [0, 1]")

        s = array_to_check.sum(axis=-1)
        if not np.all(np.isclose(s, 1)):
            log.warning(
                f"Probability array data frame may not be correctly normalised, "
                f"received probabilities do not sum to 1")

    def _predict_class_probabilities(self, x: pd.DataFrame) -> pd.DataFrame:
        return self._predict_df_through_array(x, self.get_predicted_variable_names())

    # just renaming the abstract method to implement
    def _predict_array(self, x: np.ndarray) -> np.ndarray:
        return self._predict_probabilities_array(x)

    @abstractmethod
    def _predict_probabilities_array(self, x: np.ndarray) -> np.ndarray:
        """
        If you are implementing a probabilistic classifier, this method should return a tensor with probabilities
        of shape `(N_DataPoints, N_Labels)`. It is assumed that labels are lexicographically sorted and the order
        of predictions in the output array should respect this.

        The default implementation of _predict will then use the output of this method and convert it to predicted labels (via argmax).

        In case you want to predict labels only or have a more efficient implementation of predicting labels than
        using argmax, your will have to override _predict in your implementation. In the former case of a
        non-probabilistic classifier, the implementation of this method should raise an exception, like the one below.
        """
        raise NotImplementedError(f"Model {self.__class__.__name__} does not support prediction of probabilities")

    def _predict(self, x: pd.DataFrame) -> pd.DataFrame:
        try:
            predicted_probabilities_df = self._predict_class_probabilities(x)
        except Exception:
            raise Exception(f"Wrong implementation of {self.__class__.__name__}. For non-probabilistic classifiers "
                            "_predict has to be overrode!")
        return self.convert_class_probabilities_to_predictions(predicted_probabilities_df)

    # TODO or not TODO: I don't see how to reduce the code duplication here...
    def predict(self, x: pd.DataFrame) -> pd.DataFrame:
        """
        Returns an array of integers. If the model was fitted on binary ground truth arrays of
        shape `(*tensorShape, numLabels)`, predictions will have the shape `tensorShape` and contain integers
        0, 1, ..., numLabels - 1. They correspond to the predicted labels
        """
        if not self.is_fitted():
            raise Exception(f"Calling predict with unfitted model. "
                            f"This might lead to errors down the line, especially if input/output checks are enabled")
        if self.check_input_shape:
            _check_df_shape(x, self.get_model_input_shape())
        y = super().predict(x)
        if self.check_output_shape:
            _check_df_shape(y, self.get_model_output_shape())
        return y



"""
This module contains methods and classes that facilitate evaluation of different types of models. The suggested
workflow for evaluation is to use these higher-level functionalities instead of instantiating
the evaluation classes directly.
"""
import functools
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Any, Union, Generic, TypeVar, Optional, Sequence, Callable, Set, Iterable, List, Iterator, Tuple

import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from .crossval import VectorModelCrossValidationData, VectorRegressionModelCrossValidationData, \
    VectorClassificationModelCrossValidationData, \
    VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator, VectorModelCrossValidator, VectorModelCrossValidatorParams
from .eval_stats import RegressionEvalStatsCollection, ClassificationEvalStatsCollection, RegressionEvalStatsPlotErrorDistribution, \
    RegressionEvalStatsPlotHeatmapGroundTruthPredictions, RegressionEvalStatsPlotScatterGroundTruthPredictions, \
    ClassificationEvalStatsPlotConfusionMatrix, ClassificationEvalStatsPlotPrecisionRecall, RegressionEvalStatsPlot, \
    ClassificationEvalStatsPlotProbabilityThresholdPrecisionRecall, ClassificationEvalStatsPlotProbabilityThresholdCounts, \
    Metric
from .eval_stats.eval_stats_base import EvalStats, EvalStatsCollection, EvalStatsPlot
from .eval_stats.eval_stats_classification import ClassificationEvalStats
from .eval_stats.eval_stats_regression import RegressionEvalStats
from .evaluator import VectorModelEvaluator, VectorModelEvaluationData, VectorRegressionModelEvaluator, \
    VectorRegressionModelEvaluationData, VectorClassificationModelEvaluator, VectorClassificationModelEvaluationData, \
    RegressionEvaluatorParams, ClassificationEvaluatorParams
from ..data import InputOutputData
from ..feature_importance import AggregatedFeatureImportance, FeatureImportanceProvider, plot_feature_importance, FeatureImportance
from ..tracking import TrackedExperiment
from ..tracking.tracking_base import TrackingContext
from ..util.deprecation import deprecated
from ..util.io import ResultWriter
from ..util.string import pretty_string_repr
from ..vector_model import VectorClassificationModel, VectorRegressionModel, VectorModel, VectorModelBase

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=EvalStats)
TEvalStatsPlot = TypeVar("TEvalStatsPlot", bound=EvalStatsPlot)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)
TEvaluator = TypeVar("TEvaluator", bound=VectorModelEvaluator)
TCrossValidator = TypeVar("TCrossValidator", bound=VectorModelCrossValidator)
TEvalData = TypeVar("TEvalData", bound=VectorModelEvaluationData)
TCrossValData = TypeVar("TCrossValData", bound=VectorModelCrossValidationData)


def _is_regression(model: Optional[VectorModel], is_regression: Optional[bool]) -> bool:
    if model is None and is_regression is None or (model is not None and is_regression is not None):
        raise ValueError("One of the two parameters have to be passed: model or isRegression")

    if is_regression is None:
        model: VectorModel
        return model.is_regression_model()
    return is_regression


def create_vector_model_evaluator(data: InputOutputData, model: VectorModel = None,
        is_regression: bool = None, params: Union[RegressionEvaluatorParams, ClassificationEvaluatorParams] = None,
        test_data: Optional[InputOutputData] = None) \
            -> Union[VectorRegressionModelEvaluator, VectorClassificationModelEvaluator]:
    is_regression = _is_regression(model, is_regression)
    if params is None:
        if is_regression:
            params = RegressionEvaluatorParams(fractional_split_test_fraction=0.2)
        else:
            params = ClassificationEvaluatorParams(fractional_split_test_fraction=0.2)
        log.debug(f"No evaluator parameters specified, using default: {params}")
    if is_regression:
        return VectorRegressionModelEvaluator(data, test_data=test_data, params=params)
    else:
        return VectorClassificationModelEvaluator(data, test_data=test_data, params=params)


def create_vector_model_cross_validator(data: InputOutputData,
        model: VectorModel = None,
        is_regression: bool = None,
        params: Union[VectorModelCrossValidatorParams, Dict[str, Any]] = None) \
        -> Union[VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator]:
    if params is None:
        raise ValueError("params must not be None")
    cons = VectorRegressionModelCrossValidator if _is_regression(model, is_regression) else VectorClassificationModelCrossValidator
    return cons(data, params=params)


def create_evaluation_util(data: InputOutputData, model: VectorModel = None, is_regression: bool = None,
        evaluator_params: Optional[Union[RegressionEvaluatorParams, ClassificationEvaluatorParams]] = None,
        cross_validator_params: Optional[Dict[str, Any]] = None, test_io_data: Optional[InputOutputData] = None) \
            -> Union["ClassificationModelEvaluation", "RegressionModelEvaluation"]:
    if _is_regression(model, is_regression):
        return RegressionModelEvaluation(data, evaluator_params=evaluator_params, cross_validator_params=cross_validator_params, test_io_data=test_io_data)
    else:
        return ClassificationModelEvaluation(data, evaluator_params=evaluator_params, cross_validator_params=cross_validator_params, test_io_data=test_io_data)


def eval_model_via_evaluator(model: TModel, io_data: InputOutputData, test_fraction=0.2,
        plot_target_distribution=False, compute_probabilities=True, normalize_plots=True, random_seed=60) -> TEvalData:
    """
    Evaluates the given model via a simple evaluation mechanism that uses a single split

    :param model: the model to evaluate
    :param io_data: data on which to evaluate
    :param test_fraction: the fraction of the data to test on
    :param plot_target_distribution: whether to plot the target values distribution in the entire dataset
    :param compute_probabilities: only relevant if the model is a classifier
    :param normalize_plots: whether to normalize plotted distributions such that the sum/integrate to 1
    :param random_seed:

    :return: the evaluation data
    """
    if plot_target_distribution:
        title = "Distribution of target values in entire dataset"
        fig = plt.figure(title)

        output_distribution_series = io_data.outputs.iloc[:, 0]
        log.info(f"Description of target column in training set: \n{output_distribution_series.describe()}")
        if not model.is_regression_model():
            output_distribution_series = output_distribution_series.value_counts(normalize=normalize_plots)
            ax = sns.barplot(output_distribution_series.index, output_distribution_series.values)
            ax.set_ylabel("%")
        else:
            ax = sns.distplot(output_distribution_series)
            ax.set_ylabel("Probability density")
        ax.set_title(title)
        ax.set_xlabel("target value")
        fig.show()

    if model.is_regression_model():
        evaluator_params = RegressionEvaluatorParams(fractional_split_test_fraction=test_fraction,
            fractional_split_random_seed=random_seed)
    else:
        evaluator_params = ClassificationEvaluatorParams(fractional_split_test_fraction=test_fraction,
            compute_probabilities=compute_probabilities, fractional_split_random_seed=random_seed)
    ev = create_evaluation_util(io_data, model=model, evaluator_params=evaluator_params)
    return ev.perform_simple_evaluation(model, show_plots=True, log_results=True)


class EvaluationResultCollector:
    def __init__(self, show_plots: bool = True, result_writer: Optional[ResultWriter] = None,
            tracking_context: TrackingContext = None):
        self.show_plots = show_plots
        self.result_writer = result_writer
        self.tracking_context = tracking_context

    def is_plot_creation_enabled(self) -> bool:
        return self.show_plots or self.result_writer is not None or self.tracking_context is not None

    def add_figure(self, name: str, fig: matplotlib.figure.Figure):
        if self.result_writer is not None:
            self.result_writer.write_figure(name, fig, close_figure=False)
        if self.tracking_context is not None:
            self.tracking_context.track_figure(name, fig)
        if not self.show_plots:
            plt.close(fig)

    def add_data_frame_csv_file(self, name: str, df: pd.DataFrame):
        if self.result_writer is not None:
            self.result_writer.write_data_frame_csv_file(name, df)

    def child(self, added_filename_prefix):
        result_writer = self.result_writer
        if result_writer:
            result_writer = result_writer.child_with_added_prefix(added_filename_prefix)
        return self.__class__(show_plots=self.show_plots, result_writer=result_writer)


class EvalStatsPlotCollector(Generic[TEvalStats, TEvalStatsPlot]):
    def __init__(self):
        self.plots: Dict[str, EvalStatsPlot] = {}
        self.disabled_plots: Set[str] = set()

    def add_plot(self, name: str, plot: EvalStatsPlot):
        self.plots[name] = plot

    def get_enabled_plots(self) -> List[str]:
        return [p for p in self.plots if p not in self.disabled_plots]

    def disable_plots(self, *names: str):
        self.disabled_plots.update(names)

    def create_plots(self, eval_stats: EvalStats, subtitle: str, result_collector: EvaluationResultCollector):
        known_plots = set(self.plots.keys())
        unknown_disabled_plots = self.disabled_plots.difference(known_plots)
        if len(unknown_disabled_plots) > 0:
            log.warning(f"Plots were disabled which are not registered: {unknown_disabled_plots}; known plots: {known_plots}")
        for name, plot in self.plots.items():
            if name not in self.disabled_plots and plot.is_applicable(eval_stats):
                fig = plot.create_figure(eval_stats, subtitle)
                if fig is not None:
                    result_collector.add_figure(name, fig)


class RegressionEvalStatsPlotCollector(EvalStatsPlotCollector[RegressionEvalStats, RegressionEvalStatsPlot]):
    def __init__(self):
        super().__init__()
        self.add_plot("error-dist", RegressionEvalStatsPlotErrorDistribution())
        self.add_plot("heatmap-gt-pred", RegressionEvalStatsPlotHeatmapGroundTruthPredictions(weighted=False))
        self.add_plot("heatmap-gt-pred-weighted", RegressionEvalStatsPlotHeatmapGroundTruthPredictions(weighted=True))
        self.add_plot("scatter-gt-pred", RegressionEvalStatsPlotScatterGroundTruthPredictions())


class ClassificationEvalStatsPlotCollector(EvalStatsPlotCollector[RegressionEvalStats, RegressionEvalStatsPlot]):
    def __init__(self):
        super().__init__()
        self.add_plot("confusion-matrix-rel", ClassificationEvalStatsPlotConfusionMatrix(normalise=True))
        self.add_plot("confusion-matrix-abs", ClassificationEvalStatsPlotConfusionMatrix(normalise=False))
        # the plots below apply to the binary case only (skipped for non-binary case)
        self.add_plot("precision-recall", ClassificationEvalStatsPlotPrecisionRecall())
        self.add_plot("threshold-precision-recall", ClassificationEvalStatsPlotProbabilityThresholdPrecisionRecall())
        self.add_plot("threshold-counts", ClassificationEvalStatsPlotProbabilityThresholdCounts())


class ModelEvaluation(ABC, Generic[TModel, TEvaluator, TEvalData, TCrossValidator, TCrossValData, TEvalStats]):
    """
    Utility class for the evaluation of models based on a dataset
    """
    def __init__(self, io_data: InputOutputData,
            eval_stats_plot_collector: Union[RegressionEvalStatsPlotCollector, ClassificationEvalStatsPlotCollector],
            evaluator_params: Optional[Union[RegressionEvaluatorParams, ClassificationEvaluatorParams,
                Dict[str, Any]]] = None,
            cross_validator_params: Optional[Union[VectorModelCrossValidatorParams, Dict[str, Any]]] = None,
            test_io_data: Optional[InputOutputData] = None):
        """
        :param io_data: the data set to use for evaluation. For evaluation purposes, this dataset usually will be split
            into training and test data according to the rules specified by `evaluator_params`.
            However, if `test_io_data` is specified, then this is taken to be the training data and `test_io_data` is
            taken to be the test data when creating evaluators for simple (single-split) evaluation.
        :param eval_stats_plot_collector: a collector for plots generated from evaluation stats objects
        :param evaluator_params: parameters with which to instantiate evaluators
        :param cross_validator_params: parameters with which to instantiate cross-validators
        :param test_io_data: optional test data (see `io_data`)
        """
        if cross_validator_params is None:
            cross_validator_params = VectorModelCrossValidatorParams(folds=5)
        self.evaluator_params = evaluator_params
        self.cross_validator_params = cross_validator_params
        self.io_data = io_data
        self.test_io_data = test_io_data
        self.eval_stats_plot_collector = eval_stats_plot_collector

    def create_evaluator(self, model: TModel = None, is_regression: bool = None) -> TEvaluator:
        """
        Creates an evaluator holding the current input-output data

        :param model: the model for which to create an evaluator (just for reading off regression or classification,
            the resulting evaluator will work on other models as well)
        :param is_regression: whether to create a regression model evaluator. Either this or model have to be specified
        :return: an evaluator
        """
        return create_vector_model_evaluator(self.io_data, model=model, is_regression=is_regression, test_data=self.test_io_data,
            params=self.evaluator_params)

    def create_cross_validator(self, model: TModel = None, is_regression: bool = None) -> TCrossValidator:
        """
        Creates a cross-validator holding the current input-output data

        :param model: the model for which to create a cross-validator (just for reading off regression or classification,
            the resulting evaluator will work on other models as well)
        :param is_regression: whether to create a regression model cross-validator. Either this or model have to be specified
        :return: an evaluator
        """
        return create_vector_model_cross_validator(self.io_data, model=model, is_regression=is_regression,
            params=self.cross_validator_params)

    def perform_simple_evaluation(self, model: TModel,
            create_plots=True, show_plots=False,
            log_results=True,
            result_writer: ResultWriter = None,
            additional_evaluation_on_training_data=False,
            fit_model=True, write_eval_stats=False,
            tracked_experiment: TrackedExperiment = None,
            evaluator: Optional[TEvaluator] = None) -> TEvalData:

        if show_plots and not create_plots:
            raise ValueError("showPlots=True requires createPlots=True")
        result_writer = self._result_writer_for_model(result_writer, model)
        if evaluator is None:
            evaluator = self.create_evaluator(model)
        if tracked_experiment is not None:
            evaluator.set_tracked_experiment(tracked_experiment)
        log.info(f"Evaluating {model} via {evaluator}")

        def gather_results(result_data: VectorModelEvaluationData, res_writer, subtitle_prefix=""):
            str_eval_results = ""
            for predictedVarName in result_data.predicted_var_names:
                eval_stats = result_data.get_eval_stats(predictedVarName)
                str_eval_result = str(eval_stats)
                if log_results:
                    log.info(f"{subtitle_prefix}Evaluation results for {predictedVarName}: {str_eval_result}")
                str_eval_results += predictedVarName + ": " + str_eval_result + "\n"
                if write_eval_stats and res_writer is not None:
                    res_writer.write_pickle(f"eval-stats-{predictedVarName}", eval_stats)
            str_eval_results += f"\n\n{pretty_string_repr(model)}"
            if res_writer is not None:
                res_writer.write_text_file("evaluator-results", str_eval_results)
            if create_plots:
                with TrackingContext.from_optional_experiment(tracked_experiment, model=model) as trackingContext:
                    self.create_plots(result_data, show_plots=show_plots, result_writer=res_writer,
                        subtitle_prefix=subtitle_prefix, tracking_context=trackingContext)

        eval_result_data = evaluator.eval_model(model, fit=fit_model)
        gather_results(eval_result_data, result_writer)
        if additional_evaluation_on_training_data:
            eval_result_data_train = evaluator.eval_model(model, on_training_data=True, track=False)
            additional_result_writer = result_writer.child_with_added_prefix("onTrain-") if result_writer is not None else None
            gather_results(eval_result_data_train, additional_result_writer, subtitle_prefix="[onTrain] ")
        return eval_result_data

    @staticmethod
    def _result_writer_for_model(result_writer: Optional[ResultWriter], model: TModel) -> Optional[ResultWriter]:
        if result_writer is None:
            return None
        return result_writer.child_with_added_prefix(model.get_name() + "_")

    def perform_cross_validation(self, model: TModel, show_plots=False, log_results=True, result_writer: Optional[ResultWriter] = None,
            tracked_experiment: TrackedExperiment = None, cross_validator: Optional[TCrossValidator] = None) -> TCrossValData:
        """
        Evaluates the given model via cross-validation

        :param model: the model to evaluate
        :param show_plots: whether to show plots that visualise evaluation results (combining all folds)
        :param log_results: whether to log evaluation results
        :param result_writer: a writer with which to store text files and plots. The evaluated model's name is added to each filename
            automatically
        :param tracked_experiment: a tracked experiment with which results shall be associated
        :return: cross-validation result data
        :param cross_validator: the cross-validator to apply; if None, a suitable cross-validator will be created
        """
        result_writer = self._result_writer_for_model(result_writer, model)

        if cross_validator is None:
            cross_validator = self.create_cross_validator(model)
        if tracked_experiment is not None:
            cross_validator.set_tracked_experiment(tracked_experiment)

        cross_validation_data = cross_validator.eval_model(model)

        agg_stats_by_var = {varName: cross_validation_data.get_eval_stats_collection(predicted_var_name=varName).agg_metrics_dict()
                for varName in cross_validation_data.predicted_var_names}
        df = pd.DataFrame.from_dict(agg_stats_by_var, orient="index")

        str_eval_results = df.to_string()
        if log_results:
            log.info(f"Cross-validation results:\n{str_eval_results}")
        if result_writer is not None:
            result_writer.write_text_file("crossval-results", str_eval_results)

        with TrackingContext.from_optional_experiment(tracked_experiment, model=model) as trackingContext:
            self.create_plots(cross_validation_data, show_plots=show_plots, result_writer=result_writer,
                tracking_context=trackingContext)

        return cross_validation_data

    def compare_models(self, models: Sequence[TModel], result_writer: Optional[ResultWriter] = None, use_cross_validation=False,
            fit_models=True, write_individual_results=True, sort_column: Optional[str] = None, sort_ascending: bool = True,
            sort_column_move_to_left=True,
            also_include_unsorted_results: bool = False, also_include_cross_val_global_stats: bool = False,
            visitors: Optional[Iterable["ModelComparisonVisitor"]] = None,
            write_visitor_results=False, write_csv=False,
            tracked_experiment: Optional[TrackedExperiment] = None) -> "ModelComparisonData":
        """
        Compares several models via simple evaluation or cross-validation

        :param models: the models to compare
        :param result_writer: a writer with which to store results of the comparison
        :param use_cross_validation: whether to use cross-validation in order to evaluate models; if False, use a simple evaluation
            on test data (single split)
        :param fit_models: whether to fit models before evaluating them; this can only be False if useCrossValidation=False
        :param write_individual_results: whether to write results files on each individual model (in addition to the comparison
            summary)
        :param sort_column: column/metric name by which to sort; the fact that the column names change when using cross-validation
            (aggregation function names being added) should be ignored, simply pass the (unmodified) metric name
        :param sort_ascending: whether to sort using `sortColumn` in ascending order
        :param sort_column_move_to_left: whether to move the `sortColumn` (if any) to the very left
        :param also_include_unsorted_results: whether to also include, for the case where the results are sorted, the unsorted table of
            results in the results text
        :param also_include_cross_val_global_stats: whether to also include, when using cross-validation, the evaluation metrics obtained
            when combining the predictions from all folds into a single collection. Note that for classification models,
            this may not always be possible (if the set of classes know to the model differs across folds)
        :param visitors: visitors which may process individual results
        :param write_visitor_results: whether to collect results from visitors (if any) after the comparison
        :param write_csv: whether to write metrics table to CSV files
        :param tracked_experiment: an experiment for tracking
        :return: the comparison results
        """
        # collect model evaluation results
        stats_list = []
        result_by_model_name = {}
        evaluator = None
        cross_validator = None
        for i, model in enumerate(models, start=1):
            model_name = model.get_name()
            log.info(f"Evaluating model {i}/{len(models)} named '{model_name}' ...")
            if use_cross_validation:
                if not fit_models:
                    raise ValueError("Cross-validation necessitates that models be trained several times; got fitModels=False")
                if cross_validator is None:
                    cross_validator = self.create_cross_validator(model)
                cross_val_data = self.perform_cross_validation(model, result_writer=result_writer if write_individual_results else None,
                    cross_validator=cross_validator, tracked_experiment=tracked_experiment)
                model_result = ModelComparisonData.Result(cross_validation_data=cross_val_data)
                result_by_model_name[model_name] = model_result
                eval_stats_collection = cross_val_data.get_eval_stats_collection()
                stats_dict = eval_stats_collection.agg_metrics_dict()
            else:
                if evaluator is None:
                    evaluator = self.create_evaluator(model)
                eval_data = self.perform_simple_evaluation(model, result_writer=result_writer if write_individual_results else None,
                    fit_model=fit_models, evaluator=evaluator, tracked_experiment=tracked_experiment)
                model_result = ModelComparisonData.Result(eval_data=eval_data)
                result_by_model_name[model_name] = model_result
                eval_stats = eval_data.get_eval_stats()
                stats_dict = eval_stats.metrics_dict()
            stats_dict["model_name"] = model_name
            stats_list.append(stats_dict)
            if visitors is not None:
                for visitor in visitors:
                    visitor.visit(model_name, model_result)
        results_df = pd.DataFrame(stats_list).set_index("model_name")

        # compute results data frame with combined set of data points (for cross-validation only)
        cross_val_combined_results_df = None
        if use_cross_validation and also_include_cross_val_global_stats:
            try:
                rows = []
                for model_name, result in result_by_model_name.items():
                    stats_dict = result.cross_validation_data.get_eval_stats_collection().get_global_stats().metrics_dict()
                    stats_dict["model_name"] = model_name
                    rows.append(stats_dict)
                cross_val_combined_results_df = pd.DataFrame(rows).set_index("model_name")
            except Exception as e:
                log.error(f"Creation of global stats data frame from cross-validation folds failed: {e}")

        def sorted_df(df, sort_col):
            if sort_col is not None:
                if sort_col not in df.columns:
                    alt_sort_col = f"mean[{sort_col}]"
                    if alt_sort_col in df.columns:
                        sort_col = alt_sort_col
                    else:
                        sort_col = None
                        log.warning(f"Requested sort column '{sort_col}' (or '{alt_sort_col}') not in list of columns {list(df.columns)}")
                if sort_col is not None:
                    df = df.sort_values(sort_col, ascending=sort_ascending, inplace=False)
                    if sort_column_move_to_left:
                        df = df[[sort_col] + [c for c in df.columns if c != sort_col]]
            return df

        # write comparison results
        title = "Model comparison results"
        if use_cross_validation:
            title += ", aggregated across folds"
        sorted_results_df = sorted_df(results_df, sort_column)
        str_results = f"{title}:\n{sorted_results_df.to_string()}"
        if also_include_unsorted_results and sort_column is not None:
            str_results += f"\n\n{title} (unsorted):\n{results_df.to_string()}"
        sorted_cross_val_combined_results_df = None
        if cross_val_combined_results_df is not None:
            sorted_cross_val_combined_results_df = sorted_df(cross_val_combined_results_df, sort_column)
            str_results += f"\n\nModel comparison results based on combined set of data points from all folds:\n" \
                f"{sorted_cross_val_combined_results_df.to_string()}"
        log.info(str_results)
        if result_writer is not None:
            suffix = "crossval" if use_cross_validation else "simple-eval"
            str_results += "\n\n" + "\n\n".join([f"{model.get_name()} = {model.pprints()}" for model in models])
            result_writer.write_text_file(f"model-comparison-results-{suffix}", str_results)
            if write_csv:
                result_writer.write_data_frame_csv_file(f"model-comparison-metrics-{suffix}", sorted_results_df)
                if sorted_cross_val_combined_results_df is not None:
                    result_writer.write_data_frame_csv_file(f"model-comparison-metrics-{suffix}-combined",
                        sorted_cross_val_combined_results_df)

        # write visitor results
        if visitors is not None and write_visitor_results:
            result_collector = EvaluationResultCollector(show_plots=False, result_writer=result_writer)
            for visitor in visitors:
                visitor.collect_results(result_collector)

        return ModelComparisonData(results_df, result_by_model_name, evaluator=evaluator, cross_validator=cross_validator)

    def compare_models_cross_validation(self, models: Sequence[TModel],
            result_writer: Optional[ResultWriter] = None) -> "ModelComparisonData":
        """
        Compares several models via cross-validation

        :param models: the models to compare
        :param result_writer: a writer with which to store results of the comparison
        :return: the comparison results
        """
        return self.compare_models(models, result_writer=result_writer, use_cross_validation=True)

    def create_plots(self, data: Union[TEvalData, TCrossValData], show_plots=True, result_writer: Optional[ResultWriter] = None,
            subtitle_prefix: str = "", tracking_context: Optional[TrackingContext] = None):
        """
        Creates default plots that visualise the results in the given evaluation data

        :param data: the evaluation data for which to create the default plots
        :param show_plots: whether to show plots
        :param result_writer: if not None, plots will be written using this writer
        :param subtitle_prefix: a prefix to add to the subtitle (which itself is the model name)
        :param tracking_context: the experiment tracking context
        """
        result_collector = EvaluationResultCollector(show_plots=show_plots, result_writer=result_writer,
            tracking_context=tracking_context)
        if result_collector.is_plot_creation_enabled():
            self._create_plots(data, result_collector, subtitle=subtitle_prefix + data.model_name)

    def _create_plots(self, data: Union[TEvalData, TCrossValData], result_collector: EvaluationResultCollector, subtitle=None):

        def create_plots(pred_var_name, res_collector, subt):
            if isinstance(data, VectorModelCrossValidationData):
                eval_stats = data.get_eval_stats_collection(predicted_var_name=pred_var_name).get_global_stats()
            elif isinstance(data, VectorModelEvaluationData):
                eval_stats = data.get_eval_stats(predicted_var_name=pred_var_name)
            else:
                raise ValueError(f"Unexpected argument: data={data}")
            return self._create_eval_stats_plots(eval_stats, res_collector, subtitle=subt)

        predicted_var_names = data.predicted_var_names
        if len(predicted_var_names) == 1:
            create_plots(predicted_var_names[0], result_collector, subtitle)
        else:
            for predictedVarName in predicted_var_names:
                create_plots(predictedVarName, result_collector.child(predictedVarName + "-"), f"{predictedVarName}, {subtitle}")

    def _create_eval_stats_plots(self, eval_stats: TEvalStats, result_collector: EvaluationResultCollector, subtitle=None):
        """
        :param eval_stats: the evaluation results for which to create plots
        :param result_collector: the collector to which all plots are to be passed
        :param subtitle: the subtitle to use for generated plots (if any)
        """
        self.eval_stats_plot_collector.create_plots(eval_stats, subtitle, result_collector)


class RegressionModelEvaluation(ModelEvaluation[VectorRegressionModel, VectorRegressionModelEvaluator, VectorRegressionModelEvaluationData,
        VectorRegressionModelCrossValidator, VectorRegressionModelCrossValidationData, RegressionEvalStats]):
    def __init__(self, io_data: InputOutputData,
            evaluator_params: Optional[Union[RegressionEvaluatorParams, Dict[str, Any]]] = None,
            cross_validator_params: Optional[Union[VectorModelCrossValidatorParams, Dict[str, Any]]] = None,
            test_io_data: Optional[InputOutputData] = None):
        """
        :param io_data: the data set to use for evaluation. For evaluation purposes, this dataset usually will be split
            into training and test data according to the rules specified by `evaluator_params`.
            However, if `test_io_data` is specified, then this is taken to be the training data and `test_io_data` is
            taken to be the test data when creating evaluators for simple (single-split) evaluation.
        :param evaluator_params: parameters with which to instantiate evaluators
        :param cross_validator_params: parameters with which to instantiate cross-validators
        :param test_io_data: optional test data (see `io_data`)
        """
        super().__init__(io_data, eval_stats_plot_collector=RegressionEvalStatsPlotCollector(), evaluator_params=evaluator_params,
            cross_validator_params=cross_validator_params, test_io_data=test_io_data)


class ClassificationModelEvaluation(ModelEvaluation[VectorClassificationModel, VectorClassificationModelEvaluator,
        VectorClassificationModelEvaluationData, VectorClassificationModelCrossValidator, VectorClassificationModelCrossValidationData,
        ClassificationEvalStats]):
    def __init__(self, io_data: InputOutputData,
            evaluator_params: Optional[Union[ClassificationEvaluatorParams, Dict[str, Any]]] = None,
            cross_validator_params: Optional[Union[VectorModelCrossValidatorParams, Dict[str, Any]]] = None,
            test_io_data: Optional[InputOutputData] = None):
        """
        :param io_data: the data set to use for evaluation. For evaluation purposes, this dataset usually will be split
            into training and test data according to the rules specified by `evaluator_params`.
            However, if `test_io_data` is specified, then this is taken to be the training data and `test_io_data` is
            taken to be the test data when creating evaluators for simple (single-split) evaluation.
        :param evaluator_params: parameters with which to instantiate evaluators
        :param cross_validator_params: parameters with which to instantiate cross-validators
        :param test_io_data: optional test data (see `io_data`)
        """
        super().__init__(io_data, eval_stats_plot_collector=ClassificationEvalStatsPlotCollector(), evaluator_params=evaluator_params,
            cross_validator_params=cross_validator_params, test_io_data=test_io_data)


class MultiDataModelEvaluation:
    def __init__(self, io_data_dict: Dict[str, InputOutputData], key_name: str = "dataset",
            meta_data_dict: Optional[Dict[str, Dict[str, Any]]] = None,
            evaluator_params: Optional[Union[RegressionEvaluatorParams, ClassificationEvaluatorParams, Dict[str, Any]]] = None,
            cross_validator_params: Optional[Union[VectorModelCrossValidatorParams, Dict[str, Any]]] = None,
            test_io_data_dict: Optional[Dict[str, Optional[InputOutputData]]] = None):
        """
        :param io_data_dict: a dictionary mapping from names to the data sets with which to evaluate models.
            For evaluation or cross-validation, these datasets will usually be split according to the rules
            specified by `evaluator_params or `cross_validator_params`. An exception is the case where
            explicit test data sets are specified by passing `test_io_data_dict`. Then, for these data
            sets, the io_data will not be split for evaluation, but the test_io_data will be used instead.
        :param key_name: a name for the key value used in inputOutputDataDict, which will be used as a column name in result data frames
        :param meta_data_dict: a dictionary which maps from a name (same keys as in inputOutputDataDict) to a dictionary, which maps
            from a column name to a value and which is to be used to extend the result data frames containing per-dataset results
        :param evaluator_params: parameters to use for the instantiation of evaluators (relevant if useCrossValidation==False)
        :param cross_validator_params: parameters to use for the instantiation of cross-validators (relevant if useCrossValidation==True)
        :param test_io_data_dict: a dictionary mapping from names to the test data sets to use for evaluation or to None.
            Entries with non-None values will be used for evaluation of the models that were trained on the respective io_data_dict.
            If passed, the keys need to be a superset of io_data_dict's keys (note that the values may be None, e.g.
            if you want to use test data sets for some entries, and splitting of the io_data for others).
            If not None, cross-validation cannot be used when calling ``compare_models``.
        """
        if test_io_data_dict is not None:
            missing_keys = set(io_data_dict).difference(test_io_data_dict)
            if len(missing_keys) > 0:
                raise ValueError(
                        "If test_io_data_dict is passed, its keys must be a superset of the io_data_dict's keys."
                        f"However, found missing_keys: {missing_keys}")
        self.io_data_dict = io_data_dict
        self.test_io_data_dict = test_io_data_dict

        self.key_name = key_name
        self.evaluator_params = evaluator_params
        self.cross_validator_params = cross_validator_params
        if meta_data_dict is not None:
            self.meta_df = pd.DataFrame(meta_data_dict.values(), index=meta_data_dict.keys())
        else:
            self.meta_df = None

    def compare_models(self,
            model_factories: Sequence[Callable[[], Union[VectorRegressionModel, VectorClassificationModel]]],
            use_cross_validation=False,
            result_writer: Optional[ResultWriter] = None,
            write_per_dataset_results=False,
            write_csvs=False,
            column_name_for_model_ranking: str = None,
            rank_max=True,
            add_combined_eval_stats=False,
            create_metric_distribution_plots=True,
            create_combined_eval_stats_plots=False,
            distribution_plots_cdf = True,
            distribution_plots_cdf_complementary = False,
            visitors: Optional[Iterable["ModelComparisonVisitor"]] = None) \
            -> Union["RegressionMultiDataModelComparisonData", "ClassificationMultiDataModelComparisonData"]:
        """
        :param model_factories: a sequence of factory functions for the creation of models to evaluate; every factory must result
            in a model with a fixed model name (otherwise results cannot be correctly aggregated)
        :param use_cross_validation: whether to use cross-validation (rather than a single split) for model evaluation.
            This can only be used if the instance's ``test_io_data_dict`` is None.
        :param result_writer: a writer with which to store results; if None, results are not stored
        :param write_per_dataset_results: whether to use resultWriter (if not None) in order to generate detailed results for each
            dataset in a subdirectory named according to the name of the dataset
        :param write_csvs: whether to write metrics table to CSV files
        :param column_name_for_model_ranking: column name to use for ranking models
        :param rank_max: if true, use max for ranking, else min
        :param add_combined_eval_stats: whether to also report, for each model, evaluation metrics on the combined set data points from
            all EvalStats objects.
            Note that for classification, this is only possible if all individual experiments use the same set of class labels.
        :param create_metric_distribution_plots: whether to create, for each model, plots of the distribution of each metric across the
            datasets (applies only if result_writer is not None)
        :param create_combined_eval_stats_plots: whether to combine, for each type of model, the EvalStats objects from the individual
            experiments into a single objects that holds all results and use it to create plots reflecting the overall result (applies only
            if resultWriter is not None).
            Note that for classification, this is only possible if all individual experiments use the same set of class labels.
        :param distribution_plots_cdf: whether to create CDF plots for the metric distributions. Applies only if
            create_metric_distribution_plots is True and result_writer is not None.
        :param distribution_plots_cdf_complementary: whether to plot the complementary cdf instead of the regular cdf, provided that
            distribution_plots_cdf is True.
        :param visitors: visitors which may process individual results. Plots generated by visitors are created/collected at the end of the
            comparison.
        :return: an object containing the full comparison results
        """
        if self.test_io_data_dict and use_cross_validation:
            raise ValueError("Cannot use cross-validation when `test_io_data_dict` is specified")

        all_results_df = pd.DataFrame()
        eval_stats_by_model_name = defaultdict(list)
        results_by_model_name: Dict[str, List[ModelComparisonData.Result]] = defaultdict(list)
        is_regression = None
        plot_collector: Optional[EvalStatsPlotCollector] = None
        model_names = None
        model_name_to_string_repr = None

        for i, (key, inputOutputData) in enumerate(self.io_data_dict.items(), start=1):
            log.info(f"Evaluating models for data set #{i}/{len(self.io_data_dict)}: {self.key_name}={key}")
            models = [f() for f in model_factories]

            current_model_names = [model.get_name() for model in models]
            if model_names is None:
                model_names = current_model_names
            elif model_names != current_model_names:
                log.warning(f"Model factories do not produce fixed names; use model.withName to name your models. "
                    f"Got {current_model_names}, previously got {model_names}")

            if is_regression is None:
                models_are_regression = [model.is_regression_model() for model in models]
                if all(models_are_regression):
                    is_regression = True
                elif not any(models_are_regression):
                    is_regression = False
                else:
                    raise ValueError("The models have to be either all regression models or all classification, not a mixture")

            test_io_data = self.test_io_data_dict[key] if self.test_io_data_dict is not None else None
            ev = create_evaluation_util(inputOutputData, is_regression=is_regression, evaluator_params=self.evaluator_params,
                cross_validator_params=self.cross_validator_params, test_io_data=test_io_data)

            if plot_collector is None:
                plot_collector = ev.eval_stats_plot_collector

            # compute data frame with results for current data set
            if write_per_dataset_results and result_writer is not None:
                child_result_writer = result_writer.child_for_subdirectory(key)
            else:
                child_result_writer = None
            comparison_data = ev.compare_models(models, use_cross_validation=use_cross_validation, result_writer=child_result_writer,
                visitors=visitors, write_visitor_results=False)
            df = comparison_data.results_df

            # augment data frame
            df[self.key_name] = key
            df["model_name"] = df.index
            df = df.reset_index(drop=True)

            # collect eval stats objects by model name
            for modelName, result in comparison_data.result_by_model_name.items():
                if use_cross_validation:
                    eval_stats = result.cross_validation_data.get_eval_stats_collection().get_global_stats()
                else:
                    eval_stats = result.eval_data.get_eval_stats()
                eval_stats_by_model_name[modelName].append(eval_stats)
                results_by_model_name[modelName].append(result)

            all_results_df = pd.concat((all_results_df, df))

            if model_name_to_string_repr is None:
                model_name_to_string_repr = {model.get_name(): model.pprints() for model in models}

        if self.meta_df is not None:
            all_results_df = all_results_df.join(self.meta_df, on=self.key_name, how="left")

        str_all_results = f"All results:\n{all_results_df.to_string()}"
        log.info(str_all_results)

        # create mean result by model, removing any metrics/columns that produced NaN values
        # (because the mean would be computed without them, skipna parameter unsupported)
        all_results_grouped = all_results_df.drop(columns=self.key_name).dropna(axis=1).groupby("model_name")
        mean_results_df: pd.DataFrame = all_results_grouped.mean()
        for colName in [column_name_for_model_ranking, f"mean[{column_name_for_model_ranking}]"]:
            if colName in mean_results_df:
                mean_results_df.sort_values(column_name_for_model_ranking, inplace=True, ascending=not rank_max)
                break
        str_mean_results = f"Mean results (averaged across {len(self.io_data_dict)} data sets):\n{mean_results_df.to_string()}"
        log.info(str_mean_results)

        def iter_combined_eval_stats_from_all_data_sets():
            for model_name, evalStatsList in eval_stats_by_model_name.items():
                if is_regression:
                    ev_stats = RegressionEvalStatsCollection(evalStatsList).get_global_stats()
                else:
                    ev_stats = ClassificationEvalStatsCollection(evalStatsList).get_global_stats()
                yield model_name, ev_stats

        # create further aggregations
        agg_dfs = []
        for op_name, agg_fn in [("mean", lambda x: x.mean()), ("std", lambda x: x.std()), ("min", lambda x: x.min()),
                ("max", lambda x: x.max())]:
            agg_df = agg_fn(all_results_grouped)
            agg_df.columns = [f"{op_name}[{c}]" for c in agg_df.columns]
            agg_dfs.append(agg_df)
        further_aggs_df = pd.concat(agg_dfs, axis=1)
        further_aggs_df = further_aggs_df.loc[mean_results_df.index]  # apply same sort order (index is model_name)
        column_order = functools.reduce(lambda a, b: a + b, [list(t) for t in zip(*[df.columns for df in agg_dfs])])
        further_aggs_df = further_aggs_df[column_order]
        str_further_aggs = f"Further aggregations:\n{further_aggs_df.to_string()}"
        log.info(str_further_aggs)

        # combined eval stats from all datasets (per model)
        str_combined_eval_stats = ""
        if add_combined_eval_stats:
            rows = []
            for modelName, eval_stats in iter_combined_eval_stats_from_all_data_sets():
                rows.append({"model_name": modelName, **eval_stats.metrics_dict()})
            combined_stats_df = pd.DataFrame(rows)
            combined_stats_df.set_index("model_name", drop=True, inplace=True)
            combined_stats_df = combined_stats_df.loc[mean_results_df.index]  # apply same sort order (index is model_name)
            str_combined_eval_stats = f"Results on combined test data from all data sets:\n{combined_stats_df.to_string()}\n\n"
            log.info(str_combined_eval_stats)

        if result_writer is not None:
            comparison_content = str_mean_results + "\n\n" + str_further_aggs + "\n\n" + str_combined_eval_stats + str_all_results
            comparison_content += "\n\nModels [example instance]:\n\n"
            comparison_content += "\n\n".join(f"{name} = {s}" for name, s in model_name_to_string_repr.items())
            result_writer.write_text_file("model-comparison-results", comparison_content)
            if write_csvs:
                result_writer.write_data_frame_csv_file("all-results", all_results_df)
                result_writer.write_data_frame_csv_file("mean-results", mean_results_df)

        # create plots from combined data for each model
        if create_combined_eval_stats_plots:
            for modelName, eval_stats in iter_combined_eval_stats_from_all_data_sets():
                child_result_writer = result_writer.child_with_added_prefix(modelName + "_") if result_writer is not None else None
                result_collector = EvaluationResultCollector(show_plots=False, result_writer=child_result_writer)
                plot_collector.create_plots(eval_stats, subtitle=modelName, result_collector=result_collector)

        # collect results from visitors (if any)
        result_collector = EvaluationResultCollector(show_plots=False, result_writer=result_writer)
        if visitors is not None:
            for visitor in visitors:
                visitor.collect_results(result_collector)

        # create result
        dataset_names = list(self.io_data_dict.keys())
        if is_regression:
            mdmc_data = RegressionMultiDataModelComparisonData(all_results_df, mean_results_df, further_aggs_df, eval_stats_by_model_name,
                results_by_model_name, dataset_names, model_name_to_string_repr)
        else:
            mdmc_data = ClassificationMultiDataModelComparisonData(all_results_df, mean_results_df, further_aggs_df,
                eval_stats_by_model_name, results_by_model_name, dataset_names, model_name_to_string_repr)

        # plot distributions
        if create_metric_distribution_plots and result_writer is not None:
            mdmc_data.create_distribution_plots(result_writer, cdf=distribution_plots_cdf,
                cdf_complementary=distribution_plots_cdf_complementary)

        return mdmc_data


class ModelComparisonData:
    @dataclass
    class Result:
        eval_data: Union[VectorClassificationModelEvaluationData, VectorRegressionModelEvaluationData] = None
        cross_validation_data: Union[VectorClassificationModelCrossValidationData, VectorRegressionModelCrossValidationData] = None

        def iter_evaluation_data(self) -> Iterator[Union[VectorClassificationModelEvaluationData, VectorRegressionModelEvaluationData]]:
            if self.eval_data is not None:
                yield self.eval_data
            if self.cross_validation_data is not None:
                yield from self.cross_validation_data.eval_data_list

    def __init__(self, results_df: pd.DataFrame, results_by_model_name: Dict[str, Result], evaluator: Optional[VectorModelEvaluator] = None,
            cross_validator: Optional[VectorModelCrossValidator] = None):
        self.results_df = results_df
        self.result_by_model_name = results_by_model_name
        self.evaluator = evaluator
        self.cross_validator = cross_validator

    def get_best_model_name(self, metric_name: str) -> str:
        idx = np.argmax(self.results_df[metric_name])
        return self.results_df.index[idx]

    def get_best_model(self, metric_name: str) -> Union[VectorClassificationModel, VectorRegressionModel, VectorModelBase]:
        result = self.result_by_model_name[self.get_best_model_name(metric_name)]
        if result.eval_data is None:
            raise ValueError("The best model is not well-defined when using cross-validation")
        return result.eval_data.model


class ModelComparisonVisitor(ABC):
    @abstractmethod
    def visit(self, model_name: str, result: ModelComparisonData.Result):
        pass

    @abstractmethod
    def collect_results(self, result_collector: EvaluationResultCollector) -> None:
        """
        Collects results (such as figures) at the end of the model comparison, based on the results collected

        :param result_collector: the collector to which figures are to be added
        """
        pass


class ModelComparisonVisitorAggregatedFeatureImportance(ModelComparisonVisitor):
    """
    During a model comparison, computes aggregated feature importance values for the model with the given name
    """
    def __init__(self, model_name: str, feature_agg_regex: Sequence[str] = (), write_figure=True, write_data_frame_csv=False):
        r"""
        :param model_name: the name of the model for which to compute the aggregated feature importance values
        :param feature_agg_regex: a sequence of regular expressions describing which feature names to sum as one. Each regex must
            contain exactly one group. If a regex matches a feature name, the feature importance will be summed under the key
            of the matched group instead of the full feature name. For example, the regex r"(\w+)_\d+$" will cause "foo_1" and "foo_2"
            to be summed under "foo" and similarly "bar_1" and "bar_2" to be summed under "bar".
        """
        self.model_name = model_name
        self.agg_feature_importance = AggregatedFeatureImportance(feature_agg_reg_ex=feature_agg_regex)
        self.write_figure = write_figure
        self.write_data_frame_csv = write_data_frame_csv

    def visit(self, model_name: str, result: ModelComparisonData.Result):
        if model_name == self.model_name:
            if result.cross_validation_data is not None:
                models = result.cross_validation_data.trained_models
                if models is not None:
                    for model in models:
                        self._collect(model)
                else:
                    raise ValueError("Models were not returned in cross-validation results")
            elif result.eval_data is not None:
                self._collect(result.eval_data.model)

    def _collect(self, model: Union[FeatureImportanceProvider, VectorModelBase]):
        if not isinstance(model, FeatureImportanceProvider):
            raise ValueError(f"Got model which does inherit from {FeatureImportanceProvider.__qualname__}: {model}")
        self.agg_feature_importance.add(model.get_feature_importance_dict())

    @deprecated("Use getFeatureImportance and create the plot using the returned object")
    def plot_feature_importance(self) -> plt.Figure:
        feature_importance_dict = self.agg_feature_importance.get_aggregated_feature_importance().get_feature_importance_dict()
        return plot_feature_importance(feature_importance_dict, subtitle=self.model_name)

    def get_feature_importance(self) -> FeatureImportance:
        return self.agg_feature_importance.get_aggregated_feature_importance()

    def collect_results(self, result_collector: EvaluationResultCollector):
        feature_importance = self.get_feature_importance()
        if self.write_figure:
            result_collector.add_figure(f"{self.model_name}_feature-importance", feature_importance.plot())
        if self.write_data_frame_csv:
            result_collector.add_data_frame_csv_file(f"{self.model_name}_feature-importance", feature_importance.get_data_frame())


class MultiDataModelComparisonData(Generic[TEvalStats, TEvalStatsCollection], ABC):
    def __init__(self, all_results_df: pd.DataFrame,
            mean_results_df: pd.DataFrame,
            agg_results_df: pd.DataFrame,
            eval_stats_by_model_name: Dict[str, List[TEvalStats]],
            results_by_model_name: Dict[str, List[ModelComparisonData.Result]],
            dataset_names: List[str],
            model_name_to_string_repr: Dict[str, str]):
        self.all_results_df = all_results_df
        self.mean_results_df = mean_results_df
        self.agg_results_df = agg_results_df
        self.eval_stats_by_model_name = eval_stats_by_model_name
        self.results_by_model_name = results_by_model_name
        self.dataset_names = dataset_names
        self.model_name_to_string_repr = model_name_to_string_repr

    def get_model_names(self) -> List[str]:
        return list(self.eval_stats_by_model_name.keys())

    def get_model_description(self, model_name: str) -> str:
        return self.model_name_to_string_repr[model_name]

    def get_eval_stats_list(self, model_name: str) -> List[TEvalStats]:
        return self.eval_stats_by_model_name[model_name]

    @abstractmethod
    def get_eval_stats_collection(self, model_name: str) -> TEvalStatsCollection:
        pass

    def iter_model_results(self, model_name: str) -> Iterator[Tuple[str, ModelComparisonData.Result]]:
        results = self.results_by_model_name[model_name]
        yield from zip(self.dataset_names, results)

    def create_distribution_plots(self, result_writer: ResultWriter, cdf=True, cdf_complementary=False):
        """
        Creates plots of distributions of metrics across datasets for each model as a histogram, and additionally
        any x-y plots (scatter plots & heat maps) for metrics that have associated paired metrics that were also computed

        :param result_writer: the result writer
        :param cdf: whether to additionally plot, for each distribution, the cumulative distribution function
        :param cdf_complementary: whether to plot the complementary cdf instead of the regular cdf, provided that ``cdf`` is True
        """
        for modelName in self.get_model_names():
            eval_stats_collection = self.get_eval_stats_collection(modelName)
            for metricName in eval_stats_collection.get_metric_names():
                # plot distribution
                fig = eval_stats_collection.plot_distribution(metricName, subtitle=modelName, cdf=cdf, cdf_complementary=cdf_complementary)
                result_writer.write_figure(f"{modelName}_dist-{metricName}", fig)
                # scatter plot with paired metrics
                metric: Metric = eval_stats_collection.get_metric_by_name(metricName)
                for paired_metric in metric.get_paired_metrics():
                    if eval_stats_collection.has_metric(paired_metric):
                        fig = eval_stats_collection.plot_scatter(metric.name, paired_metric.name)
                        result_writer.write_figure(f"{modelName}_scatter-{metric.name}-{paired_metric.name}", fig)
                        fig = eval_stats_collection.plot_heat_map(metric.name, paired_metric.name)
                        result_writer.write_figure(f"{modelName}_heatmap-{metric.name}-{paired_metric.name}", fig)


class ClassificationMultiDataModelComparisonData(MultiDataModelComparisonData[ClassificationEvalStats, ClassificationEvalStatsCollection]):
    def get_eval_stats_collection(self, model_name: str):
        return ClassificationEvalStatsCollection(self.get_eval_stats_list(model_name))


class RegressionMultiDataModelComparisonData(MultiDataModelComparisonData[RegressionEvalStats, RegressionEvalStatsCollection]):
    def get_eval_stats_collection(self, model_name: str):
        return RegressionEvalStatsCollection(self.get_eval_stats_list(model_name))



from .crossval import VectorClassificationModelCrossValidator, VectorRegressionModelCrossValidator, \
    VectorClassificationModelCrossValidationData, VectorRegressionModelCrossValidationData, \
    VectorModelCrossValidatorParams
from .eval_util import RegressionModelEvaluation, ClassificationModelEvaluation, MultiDataModelEvaluation, \
    eval_model_via_evaluator, create_evaluation_util, create_vector_model_evaluator, create_vector_model_cross_validator
from .evaluator import VectorClassificationModelEvaluator, VectorRegressionModelEvaluator, \
    RegressionEvaluatorParams, ClassificationEvaluatorParams, \
    VectorRegressionModelEvaluatorParams, VectorClassificationModelEvaluatorParams, \
    VectorRegressionModelEvaluationData, VectorClassificationModelEvaluationData, \
    RuleBasedVectorClassificationModelEvaluator, RuleBasedVectorRegressionModelEvaluator

from ..util import mark_used

# imports required for backward compatibility
from ..data import DataSplitter, DataSplitterFractional
mark_used(DataSplitter, DataSplitterFractional)


import collections
import logging
import math
import random
import time
from abc import ABC, abstractmethod
from typing import Optional, Tuple, Callable, Type, Sequence, TypeVar, Generic

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from .util.aggregation import RelativeFrequencyCounter

log = logging.getLogger(__name__)


class SATemperatureSchedule(ABC):
    """
    Describes how temperature changes as the annealing process goes on.
    The function maps a degree of completion (in the interval from 0 to 1) to a temperature value.
    """

    @abstractmethod
    def temperature(self, degree_of_completion: float) -> float:
        """
        Computes the temperature for a degree of completion in [0, 1]
        :param degree_of_completion: the degree to which the simulated annealing process has completed in [0, 1]
        :return: the temperature
        """
        pass

    def probability(self, degree_of_completion: float, cost_delta: float) -> Tuple[float, float]:
        T = self.temperature(degree_of_completion)
        p = math.exp(-cost_delta / T) if T > 0.0 else 0.0
        return p, T

    @abstractmethod
    def _get_params(self):
        pass

    def __str__(self):
        return f"{self.__class__.__name__}{self._get_params()}"


class SATemperatureScheduleFixed(SATemperatureSchedule):
    """A schedule with a constant temperature"""
    def __init__(self, t):
        self.t = t

    def temperature(self, degree_of_completion):
        return self.t

    def _get_params(self):
        return {"fixedTemp": self.t}


class SATemperatureScheduleExponential(SATemperatureSchedule):
    """ A temperature schedule for simulated annealing where the temperature drops exponentially. """

    def __init__(self, t0, t1, exponent_factor):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param exponent_factor: factor with which to multiply the exponent (the larger the factor, the faster the temperature drops)
        """
        super().__init__()
        if t1 > t0 or t0 <= 0:
            raise ValueError("Inadmissible temperatures given.")

        if exponent_factor < 1:
            raise ValueError("The exponent factor cannot be less than 1.")

        self.t0 = t0
        self.t1 = t1
        self.t_drop = t0 - t1
        self.exponent_factor = exponent_factor

    def temperature(self, degree_of_completion):
        return self.t0 - self.t_drop * (1.0 - math.exp(-self.exponent_factor * degree_of_completion))

    def _get_params(self):
        return {"t0": self.t0, "t1": self.t1, "exp_factor": self.exponent_factor}


def reverse_sigmoid(degree_of_completion, v0, v1, steepness, mid_degree=0.5):
    return v0 - (v0-v1) / (1.0 + math.exp(-steepness * (degree_of_completion - mid_degree)))


class SATemperatureScheduleReverseSigmoid(SATemperatureSchedule):
    """
    A temperature schedule for simulated annealing where the temperature drops in a reverse sigmoid curve (based on the logistic function)
    """

    def __init__(self, t0, t1, steepness, mid_degree=0.5):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param steepness: Factor which roughly corresponds to the (negative) slope at the inflection point 0.5.
            For the sigmoid shape to be sufficiently pronounced, a value of at least 5 is recommended (the closer the value is to 1,
            the more the curve approaches a linear decay).
        :param mid_degree: the degree of completion at which the function shall return the temperature (t0+t1)/2
        """
        super().__init__()
        if t1 > t0 or t0 <= 0:
            raise ValueError("Inadmissible temperatures given.")
        if mid_degree < 0 or mid_degree > 1:
            raise Exception("Mid-degree must be between 0 and 1")

        self.t0 = t0
        self.t1 = t1
        self.steepness = steepness
        self.mid_degree = mid_degree

    def temperature(self, degree_of_completion):
        return reverse_sigmoid(degree_of_completion, self.t0, self.t1, self.steepness, self.mid_degree)

    def _get_params(self):
        return {"t0": self.t0, "t1": self.t1, "steepness": self.steepness, "mid_degree": self.mid_degree}


class SATemperatureScheduleReverseSigmoidSymmetric(SATemperatureScheduleReverseSigmoid):
    """
    A variant of the logistic schedule with a reverse sigmoid shape, where the probability of acceptance for a given assumed positive
    cost delta is symmetric. "Symmetric" here means that half the schedule is to be above 0.5 and half of it below 0.5.
    """
    def __init__(self, t0, t1, steepness, cost_delta_for_symmetry):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param steepness: Factor which roughly corresponds to the (negative) slope at the inflection point 0.5.
            For the sigmoid shape to be sufficiently pronounced, a value of at least 8 is recommended (the closer the value is to 1,
            the more the curve approaches a linear decay).
        :param cost_delta_for_symmetry: the (positive) cost delta value which the curve shall be "symmetric"
        """
        k = steepness
        cdelta = cost_delta_for_symmetry
        mid_degree = -(2*math.log(-(13614799*t0-19642003*cdelta)/(13614799*t1-19642003*cdelta))-k)/(2*k)
        super().__init__(t0, t1, steepness, mid_degree)


class SATemperatureSchedulePower(SATemperatureSchedule):
    """
    A temperature schedule for simulated annealing where the temperature drops with powers of the degree of completion d,
    i.e. the temperature drop is proportional to d to the power of e for a given exponent e.
    """

    def __init__(self, t0, t1, exponent):
        """
        :param t0: temperature at the beginning (degree of completion 0)
        :param t1: temperature at the end (degree of completion 1), which must be smaller than t0 and larger than 0
        :param exponent: the exponent of the power function
        """
        super().__init__()
        if t1 > t0 or t0 <= 0:
            raise ValueError("Inadmissible temperatures given.")

        self.t0 = t0
        self.t1 = t1
        self.tDrop = t0 - t1
        self.exponent = exponent

    def temperature(self, degree_of_completion):
        return self.t0 - self.tDrop * (math.pow(degree_of_completion, self.exponent))

    def _get_params(self):
        return {"t0": self.t0, "t1": self.t1, "exponent": self.exponent}


class SAProbabilityFunction(ABC):
    def __init__(self, **params_dict):
        self.params_dict = params_dict

    def __str__(self):
        return f"{self.__class__.__name__}{self.params_dict}"

    @abstractmethod
    def __call__(self, degree_of_completion):
        pass


class SAProbabilityFunctionLinear(SAProbabilityFunction):
    """A probability function where probabilities decay linearly"""
    def __init__(self, p0=1.0, p1=0.0):
        super().__init__(p0=p0, p1=p1)
        self.p0 = p0
        self.p1 = p1

    def __call__(self, degree_of_completion):
        return self.p0 - degree_of_completion * (self.p0 - self.p1)


class SAProbabilityFunctionReverseSigmoid(SAProbabilityFunction):
    """A probability function where probabilities decay in a reverse sigmoid shape"""
    def __init__(self, p0=1.0, p1=0.0, steepness=10):
        """
        :param p0: the probability at the beginning (degree of completion 0)
        :param p1: the probability at the end (degree of completion 1)
        :param steepness: the steepness of the sigmoid curve
        """
        super().__init__(p0=p0, p1=p1, steepness=steepness)
        self.p0 = p0
        self.p1 = p1
        self.steepness = steepness

    def __call__(self, degree_of_completion):
        return reverse_sigmoid(degree_of_completion, self.p0, self.p1, self.steepness)


class SAProbabilityFunctionConstant(SAProbabilityFunction):
    """A constant probability function (which returns the same probability for all degrees of completion)"""
    def __init__(self, p):
        super().__init__(p=p)
        self.p = p

    def __call__(self, degree_of_completion):
        return self.p


class SAProbabilitySchedule(SATemperatureSchedule):
    """
    A temperature schedule where temperatures are derived from a probability schedule that is to apply to a reference cost delta, which
    is either given or is computed from observed values (the latter resulting in an adaptive schedule).
    It converts a function that returns probabilities for degrees of completion into a corresponding temperature schedule.
    """
    def __init__(self, reference_cost_delta: Optional[float], probability_function: SAProbabilityFunction):
        """
        Creates a temperature schedule for a reference cost delta (which can also be computed automatically from observed data)
        and probability function:
        The schedule will return temperatures such that for referenceCostDelta, the probability of acceptance of a move at
        degree of completion d in [0,1] will be probabilityFunction(d).

        :param reference_cost_delta: the (positive) cost delta for which the probability function is to apply;
            if None, adaptively determine it from the empirical mean
        :param probability_function: a function which maps degrees of completion in [0,1] to probabilities in [0,1]
        """
        self.adaptive = reference_cost_delta is None
        self.referenceCostDelta = reference_cost_delta
        self.probabilityFunction = probability_function
        self.paramsDict = {
            "refCostDelta": reference_cost_delta if reference_cost_delta is not None else "adaptive",
            "probabilityFunction": str(probability_function)}

        # helper variables for adaptive mode
        self._costDeltaSum = 0
        self._costDeltaCount = 0

    def temperature(self, degree_of_completion):
        if self.adaptive and self._costDeltaCount == 0:
            raise Exception("Cannot generate a temperature from an adaptive schedule without any previous cost-delta samples")
        p = self.probabilityFunction(degree_of_completion)
        if p == 0.0:
            return 0
        else:
            return -self.referenceCostDelta / math.log(p)

    def probability(self, degree_of_completion, cost_delta):
        if self.adaptive:
            self._costDeltaSum += cost_delta
            self._costDeltaCount += 1
            self.referenceCostDelta = max(self._costDeltaSum / self._costDeltaCount, 1e-10)
        return super().probability(degree_of_completion, cost_delta)

    def _get_params(self):
        return self.paramsDict


class SACostValue(ABC):
    """Representation of an immutable cost value"""

    @abstractmethod
    def value(self):
        """Returns the numeric cost value"""
        pass

    @abstractmethod
    def add(self, other) -> 'SACostValue':
        pass


class SACostValueNumeric(SACostValue):
    def __init__(self, scalar):
        self._value = scalar

    def __str__(self):
        return str(self.value())

    def value(self):
        return self._value

    def add(self, other: 'SACostValueNumeric') -> 'SACostValueNumeric':
        return SACostValueNumeric(self._value + other._value)


class SAState(ABC):
    """Represents the state/variable assignment during a simulated annealing process"""

    def __init__(self, r: random.Random):
        self.r = r
        self.cost = self.compute_cost_value()

    @abstractmethod
    def compute_cost_value(self) -> SACostValue:
        """Computes the costs of this state (from scratch)"""
        pass

    @abstractmethod
    def get_state_representation(self):
        """
        Returns a compact state representation (for the purpose of archiving a hitherto best result), which can later be
        applied via applyStateRepresentation.

        :return: a compact state representation of some sort
        """
        pass

    @abstractmethod
    def apply_state_representation(self, representation):
        """
        Applies the given state representation (as returned via `getStateRepresentation`) in order for the optimisation result to
        be obtained by the user.
        Note that the function does not necessarily need to change this state to reflect the representation, as its sole
        purpose is for the optimsation result to be obtainable thereafter (it is not used during the optimisation process as such).

        :param representation: a representation as returned by `getStateRepresentation`
        """
        pass


TSAState = TypeVar("TSAState", bound=SAState)


class SAOperator(Generic[TSAState]):
    """
    An operator which, when applied with appropriately chosen parameters, can transform a state into another
    state during simulated annealing
    """

    def __init__(self, state: TSAState):
        """
        :param state: the state to which the operator is applied
        """
        self.state = state

    def apply_cost_change(self, cost_delta: SACostValue):
        """
        Applies the cost change to the state given at construction

        :param cost_delta: the cost change to apply
        """
        self.state.cost = self.state.cost.add(cost_delta)

    @abstractmethod
    def apply_state_change(self, *params):
        """
        Applies the operator to the state, i.e. it makes the changes to the state only
        (and does not update the associated costs)

        :param params: the parameters with which the operator is to be applied
        :return:
        """
        pass

    def apply(self, params: Tuple, cost_delta: SACostValue):
        """
        Applies the operator to the state given at construction, changing the state and updating the costs appropriately

        :param params: the parameters with which the operator is to be applied
        :param cost_delta: the cost change that results from the application
        :return:
        """
        self.apply_cost_change(cost_delta)
        self.apply_state_change(*params)

    @abstractmethod
    def cost_delta(self, *params) -> SACostValue:
        """
        Computes the cost change that would apply when applying this operator with the given parameters

        :param params: an arbitrary list of parameters (specific to the concrete operator)
        :return:
        """
        pass

    @abstractmethod
    def choose_params(self) -> Optional[Tuple[Tuple, Optional[SACostValue]]]:
        """
        Chooses parameters for the application of the operator (e.g. randomly or greedily).

        :return: a tuple (params, costValue) or None if no suitable parameters are found, where params is the list of chosen
            parameters and costValue is either an instance of CostValue or None if the costs have not been computed.
        """
        pass


class SAChain(Generic[TSAState]):
    """Manages the progression of one state during simulated annealing"""

    log = log.getChild(__qualname__)

    def __init__(self,
            state_factory: Callable[[random.Random], TSAState],
            schedule: SATemperatureSchedule,
            ops_and_weights: Sequence[Tuple[Callable[[TSAState], SAOperator[TSAState]], float]],
            random_seed,
            collect_stats=False):
        self.schedule = schedule
        self.r = random.Random(random_seed)
        self.state = state_factory(self.r)
        self.collect_stats = collect_stats
        operators, weights = zip(*ops_and_weights)
        cum_weights, s = [], 0
        for weight in weights:
            s += weight
            cum_weights.append(s)
        self.ops = [cons(self.state) for cons in operators]
        self.op_cum_weights = cum_weights
        self.steps_taken = 0
        self.count_none_params = 0
        self.count_best_updates = -1
        self.best_cost = None
        self.best_state_repr = None
        self.logged_series = collections.defaultdict(lambda: [])
        self._update_best_state()

        if self.collect_stats:
            self.operator_inapplicability_counters = {}
            for op in self.ops:
                self.operator_inapplicability_counters[op] = RelativeFrequencyCounter()

    def _update_best_state(self):
        cost = self.state.cost
        if self.best_cost is None or cost.value() < self.best_cost.value():
            self.best_cost = cost
            self.best_state_repr = self.state.get_state_representation()
            self.count_best_updates += 1

    def step(self, degree_of_completion):
        r = self.r

        # make move
        op = r.choices(self.ops, cum_weights=self.op_cum_weights, k=1)[0]
        param_choice = op.choose_params()
        if param_choice is None:
            self.count_none_params += 1
        else:
            params, cost_change = param_choice
            if cost_change is None:
                cost_change = op.cost_delta(*params)
            if cost_change.value() < 0:
                make_move = True
            else:
                cost_change_value = cost_change.value()
                p, T = self.schedule.probability(degree_of_completion, cost_change_value)
                make_move = r.random() <= p
                self.log.debug(f'p: {p}, T: {T}, costDelta: {cost_change_value}, move: {make_move}')
                if self.collect_stats:
                    self.logged_series["temperatures"].append(T)
                    self.logged_series["probabilities"].append(p)
            if make_move:
                op.apply(params, cost_change)
                self._update_best_state()
            if self.collect_stats:
                self.logged_series["costDeltas"].append(cost_change.value())
        if self.collect_stats:
            self.logged_series["bestCostValues"].append(self.best_cost.value())
            self.logged_series["costValues"].append(self.state.cost.value())
            self.operator_inapplicability_counters[op].count(param_choice is None)

        self.steps_taken += 1

        if self.log.isEnabledFor(logging.DEBUG):
            self.log.debug(f"Step {self.steps_taken}: cost={self.state.cost}; best cost={self.best_cost}")

    def log_stats(self):
        if self.collect_stats:
            stats = {"useless moves total (None params)": f"{self.count_none_params}/{self.steps_taken}"}
            for op, counter in self.operator_inapplicability_counters.items():
                stats[f"useless moves of {op}"] = str(counter)
            logged_cost_deltas = self.logged_series["costDeltas"]
            if logged_cost_deltas:
                stats["mean cost delta"] = f"{np.mean(logged_cost_deltas):.3f} +- { np.std(logged_cost_deltas):.3f}"
                abs_cost_deltas = np.abs(logged_cost_deltas)
                stats["mean absolute cost delta"] = f"{np.mean(abs_cost_deltas):.3f} +- {np.std(abs_cost_deltas):.3f}"
                positive_cost_deltas = [cd for cd in logged_cost_deltas if cd > 0]
                if positive_cost_deltas:
                    stats["positive cost delta"] = f"mean={np.mean(positive_cost_deltas):.3f} +- {np.std(positive_cost_deltas):.3f}," \
                                                   f" max={np.max(positive_cost_deltas):.3f}"
            stats_join = "\n    "
            self.log.info(f"Stats: {stats_join.join([key + ': ' + value for (key, value) in stats.items()])}")
        self.log.info(f"Best solution has {self.best_cost} after {self.count_best_updates} updates of best state")

    def apply_best_state(self):
        """Applies the best state representation found in this chain to the chain's state"""
        self.state.apply_state_representation(self.best_state_repr)
        self.state.cost = self.best_cost

    def plot_series(self, series_name):
        """
        Plots one of the logged series

        :param series_name: the name of the series (see getSeries)
        """
        series = self.get_series(series_name)
        plt.figure()
        series.plot(title=series_name)

    def get_series(self, series_name):
        """
        Gets one of the logged series (for collectStats==True)

        :param series_name: name of the series: one of "temperatures", "probabilities", "costDeltas", "bestCostValues", "costValues
        """
        if not self.collect_stats:
            raise Exception("No stats were collected")
        if series_name not in self.logged_series:
            raise Exception("Unknown series")
        return pd.Series(self.logged_series[series_name])


class SimulatedAnnealing(Generic[TSAState]):
    """
    The simulated annealing algorithm for discrete optimisation (cost minimisation)
    """

    log = log.getChild(__qualname__)

    def __init__(self,
            schedule_factory: Callable[[], SATemperatureSchedule],
            ops_and_weights: Sequence[Tuple[Callable[[TSAState], SAOperator[TSAState]], float]],
            max_steps: int = None,
            duration: float = None,
            random_seed=42,
            collect_stats=False):
        """
        :param schedule_factory: a factory for the creation of the temperature schedule for the annealing process
        :param ops_and_weights: a list of operator factories with associated weights, where weights are to indicate the (non-normalised)
            probability of choosing the associated operator
        :param max_steps: the number of steps for which to run the optimisation; may be None (if not given, duration must be provided)
        :param duration: the duration, in seconds, for which to run the optimisation; may be None (if not given, maxSteps must be provided)
        :param random_seed: the random seed to use for all random choices
        :param collect_stats: flag indicating whether to collect additional statics which will be logged
        """
        if max_steps is not None and max_steps <= 0:
            raise ValueError("The number of iterations should be greater than 0.")
        if max_steps is None and duration is None or (max_steps is not None and duration is not None):
            raise ValueError("Exactly one of {maxSteps, duration} must be specified.")
        if duration is not None and duration <= 0:
            raise ValueError("Duration must be greater than 0 if provided")
        self.scheduleFactory = schedule_factory
        self.max_steps = max_steps
        self.duration = duration
        self.randomSeed = random_seed
        self.opsAndWeights = ops_and_weights
        self.collect_stats = collect_stats
        self._chain = None

    def optimise(self, state_factory: Callable[[random.Random], TSAState]) -> TSAState:
        """
        Applies the annealing process, starting with a state created via the given factory.

        :param state_factory: the factory with which to create the (initial) state
        :return: the state with the least-cost representation found during the optimisation applied
        """
        chain = SAChain(state_factory, self.scheduleFactory(), ops_and_weights=self.opsAndWeights, random_seed=self.randomSeed,
            collect_stats=self.collect_stats)
        self.log.info(f"Running simulated annealing with {len(self.opsAndWeights)} operators for "
                      f"{'%d steps' % self.max_steps if self.max_steps is not None else '%d seconds' % self.duration} ...")
        start_time = time.time()
        while True:
            time_elapsed = time.time() - start_time
            if (self.max_steps is not None and chain.steps_taken >= self.max_steps) or (self.duration is not None and time_elapsed >= self.duration):
                break
            if self.max_steps is not None:
                degree_of_completion = chain.steps_taken / self.max_steps
            else:
                degree_of_completion = time_elapsed / self.duration
            chain.step(degree_of_completion)
        self.log.info(f"Simulated annealing completed after {time.time()-start_time:.1f} seconds, {chain.steps_taken} steps")
        chain.log_stats()
        chain.apply_best_state()
        if self.collect_stats:
            self._chain = chain
        return chain.state

    def get_chain(self) -> Optional[SAChain[TSAState]]:
        """
        Gets the chain used by the most recently completed application (optimise call) of this object
        for the case where stats collection was enabled; the chain then contains relevant series and may be used
        to generate plots. If stats collection was not enabled, returns None.
        """
        return self._chain


class ParallelTempering(Generic[TSAState]):
    """
    The parallel tempering algorithm for discrete optimisation (cost minimisation)
    """

    log = log.getChild(__qualname__)

    def __init__(self,
            num_chains,
            ops_and_weights: Sequence[Tuple[Callable[[TSAState], SAOperator[TSAState]], float]],
            schedule: SATemperatureSchedule = None,
            probability_function: SAProbabilityFunction = None,
            max_steps: int = None,
            duration: float = None,
            random_seed=42,
            log_cost_progression=False):
        """
        Creates a parallel tempering optimiser with the given number of chains and operators for each chain.
        To determine the schedule to use for each chain, either schedule or probabilityFunction must be provided.
        It is usually more robust to use adaptive schedules and therefore to provide probabilityFunction.

        :param num_chains: the number of parallel chains
        :param ops_and_weights: a list of operators with associated weights (which are to indicate the non-normalised probability of
            choosing the associated operator)
        :param schedule: the temperature schedule from which numChains temperatures of chains are drawn (using equidistant degrees of
            completion); if None, must provide probabilityFunction
        :param probability_function: the probability function from which numChains probabilities for adaptive probability schedules, each
            using a constant probability, are to be drawn; if None, must provide schedule
        :param max_steps: the number of steps for which to run the optimisation; may be None (if not given, duration must be provided)
        :param duration: the duration, in seconds, for which to run the optimisation; may be None (if not given, maxSteps must be provided)
        :param random_seed: the random seed to use for all random choices
        :param log_cost_progression: whether to log cost progression of all chains (such that it can be plotted after the fact via
            plotCostProgression)
        """
        if max_steps is not None and max_steps <= 0:
            raise ValueError("The number of iterations should be greater than 0.")
        if (max_steps is None and duration is None) or (max_steps is not None and duration is not None):
            raise ValueError("Exactly one of {maxSteps, duration} must be specified.")
        if duration is not None and duration <= 0:
            raise ValueError("duration should be greater than 0 if provided.")
        if num_chains < 2:
            raise ValueError("Number of chains must be at least 2.")
        if (schedule is None and probability_function is None) or (schedule is not None and probability_function is not None):
            raise ValueError("Exactly one of {schedule, probabilityFunction} must be given")
        self.max_steps = max_steps
        self.duration = duration
        self.random_seed = random_seed
        self.num_chains = num_chains
        self.base_schedule = schedule
        self.base_probability_function = probability_function
        self.ops_and_weights = ops_and_weights
        self.log_cost_progression = log_cost_progression

        # transient members
        self._cost_progressions = None
        self._schedule_param_strings = None

    def _create_schedules(self):
        degree_step = 1.0 / (self.num_chains - 1)
        degrees_of_completion = [i*degree_step for i in range(self.num_chains)]
        if self.base_schedule is not None:
            # create schedules with fixed temperatures taken from base schedule
            temperatures = [self.base_schedule.temperature(d) for d in degrees_of_completion]
            self._schedule_param_strings = ["T=%.2f" % t for t in temperatures]
            return [SATemperatureScheduleFixed(t) for t in temperatures]
        else:
            # create adaptive probability schedules based on probabilities taken from base probability function
            probabilities = [self.base_probability_function(d) for d in degrees_of_completion]
            self._schedule_param_strings = ["p=%.3f" % p for p in probabilities]
            return [SAProbabilitySchedule(None, SAProbabilityFunctionConstant(p)) for p in probabilities]

    def optimise(self, state_factory: Callable[[random.Random], SAState]) -> SAState:
        """
        Applies the optimisation process, starting, in each chain, with a state created via the given factory.

        :param state_factory: the factory with which to create the states for all chains
        :return: the state with the least-cost representation found during the optimisation (among all parallel chains) applied
        """
        self.log.info(f"Running parallel tempering with {self.num_chains} chains, {len(self.ops_and_weights)} operators for "
                      f"{'%d steps' % self.max_steps if self.max_steps is not None else '%d seconds' % self.duration} ...")

        r = random.Random(self.random_seed)
        chains = []
        cost_progressions = []
        for i, schedule in enumerate(self._create_schedules(), start=1):
            self.log.info(f"Chain {i} uses {schedule}")
            chains.append(SAChain(state_factory, schedule, ops_and_weights=self.ops_and_weights, random_seed=r.randint(0, 1000)))
            cost_progressions.append([])

        start_time = time.time()
        step = 0
        num_chain_swaps = 0
        while True:
            time_elapsed = time.time() - start_time
            if (self.max_steps is not None and step > self.max_steps) or (self.duration is not None and time_elapsed > self.duration):
                break

            # take one step in each chain
            degree_of_completion = step / self.max_steps if self.max_steps is not None else time_elapsed / self.duration
            for chain in chains:
                chain.step(degree_of_completion)

            # check if neighbouring chains can be "swapped": if a high-temperature chain has a better state
            # than a low-temperature chain, swap them (by exchanging their schedules and swapping them
            # in the chains array, which shall always be in descending order of temperature)
            for idx_high_temp_chain in range(0, self.num_chains - 1):
                idx_low_temp_chain = idx_high_temp_chain+1
                high_temp_chain = chains[idx_high_temp_chain]
                low_temp_chain = chains[idx_low_temp_chain]
                if high_temp_chain.state.cost.value() < low_temp_chain.state.cost.value():
                    high_temp_chain.schedule, low_temp_chain.schedule = low_temp_chain.schedule, high_temp_chain.schedule
                    chains[idx_low_temp_chain] = high_temp_chain
                    chains[idx_high_temp_chain] = low_temp_chain
                    num_chain_swaps += 1

            if self.log_cost_progression:
                for idx_chain, chain in enumerate(chains):
                    cost_progressions[idx_chain].append(chain.state.cost.value())

            step += 1

        self.log.info(f"Number of chain swaps: {num_chain_swaps}")
        if self.log_cost_progression: self._cost_progressions = cost_progressions

        # apply best solution
        best_chain_idx = int(np.argmin([chain.best_cost.value() for chain in chains]))
        chains[best_chain_idx].apply_best_state()
        return chains[best_chain_idx].state

    def plot_cost_progression(self):
        if not self.log_cost_progression or self._cost_progressions is None:
            raise Exception("No cost progression was logged")
        series = {}
        for scheduleParamStr, costProgression in zip(self._schedule_param_strings, self._cost_progressions):
            series[scheduleParamStr] = costProgression
        plt.figure()
        pd.DataFrame(series).plot()
